import unittest
from particleShear import *
from tkinter import *
import math



class TesttwoSpheresGeneral(unittest.TestCase):

    def setUp(self):



        size_x = 500
        size_y = 500

        theTk = Tk()
        theCanvas = Canvas(theTk, width=size_x, height=size_y)

        theCanvas.pack()

        theEnsemble = EnsembleLinkable(500, 500, 0, 0.8, theCanvas, True, k=1, nu=1, k_t=1, nu_t=1, mu=0.01)

        theEnsemble.sphereList.append(SphereLinkable("GREY", 150, 200, 120, m=1, my_index=1, theCanvas=theCanvas,
                                                     doDrawing=True, force_register=theEnsemble,
                                                     size_x=theEnsemble.size_x,
                                                     size_y=theEnsemble.size_y))

        theEnsemble.sphereList[0].xspeed = 10

        theEnsemble.sphereList.append(SphereLinkable("BLUE", 220, 260, 120, m=1, my_index=2, theCanvas=theCanvas,
                                                     doDrawing=True, force_register=theEnsemble,
                                                     size_x=theEnsemble.size_x,
                                                     size_y=theEnsemble.size_y))

        theEnsemble.sphereList[1].xspeed = -10

        theEnsemble.sphereList[1].omega = 0.25

        graphical = theEnsemble.graphical_output_configuration

        graphical.draw_rotation_line = True

        graphical.draw_permanent_connections=True

        theEnsemble.set_graphical_output_configuration(graphical)

        theEnsemble.sphereList[0].establish_permanent_link(theEnsemble.sphereList[1])

        theTk.update()

        theEvaluator = StressTensorEvaluation(size_x, size_y)

        self.theEvaluator = theEvaluator
        self.theEnsemble = theEnsemble
        self.theTk=theTk



        self.dt=0.01
        self.cool_factor=1

        for i in range(50):
            self.theEnsemble.mechanical_simulation_step_calculate_forces()

            self.theEnsemble.record_total_particle_forces()
            self.theEvaluator.evaluate_stress_tensors(self.theEnsemble.force_register, self.theEnsemble.sphereList, 0)


            self.theEnsemble.mechanical_simulation_step_calculate_acceleration(cool_factor=self.cool_factor, dt=self.dt)
            self.theEnsemble.mechanical_simulation_step_calculate_movement(dt=self.dt)
            self.theTk.update()


    def tearDown(self):
        self.theTk.destroy()
        self.theTk = False





    def test_stress_tensor_LW(self):






        LW = self.theEvaluator.stress_tensor_LW





        self.assertTrue(LW[0][0]>0 and LW[1][1]>0)



    def test_overall_stress_tensor(self):



        overall = self.theEvaluator.overall_stress_tensor

        overall_tot = abs(overall[0][0])+abs(overall[0][1])+abs(overall[1][0])+abs(overall[1][1])

        self.assertTrue(abs((overall[0][1] - overall[1][0]) / overall_tot) < 1e-15)

        self.assertTrue(overall[0][0] > 0 and overall[1][1] > 0)

    def test_unbalanced_forces_stress_tensor(self):




        unbalanced = self.theEvaluator.stress_tensor_unbalanced_forces

        LW = self.theEvaluator.stress_tensor_LW


        LW_tot=abs(LW[0][0])+abs(LW[0][1])+abs(LW[1][0])+abs(LW[1][1])

        self.assertTrue(abs(LW[0][0] - unbalanced[0][0])/LW_tot < 1e-15 and
                        abs(LW[0][1] - unbalanced[0][1]) / LW_tot < 1e-15 and
                        abs(LW[1][0] - unbalanced[1][0]) / LW_tot < 1e-15 and
                        abs(LW[1][1] - unbalanced[1][1]) / LW_tot < 1e-15
                        )


    def test_external_stress_tensor(self):


        external = self.theEvaluator.stress_tensor_with_external_forces
        self.assertTrue(
            external[0][0] == 0 and external[0][1] == 0 and external[1][0] == 0 and external[1][1] == 0)

    def test_peculiar_acceleration_stress_tensor(self):


        peculiar = self.theEvaluator.stress_tensor_linear_acceleration_otsuki



    def test_stress_tensor_spin_kinetic_energy(self):

        stress_tensor_spin = self.theEvaluator.stress_tensor_spin_kinetic_energy

        # In theory, the spin kinetic energy should be I omega*w^2

        self.assertTrue(
            stress_tensor_spin[0][1] == stress_tensor_spin[1][0] )

    def test_stress_tensor_unbalanced_torque(self):

        for i in range(20):
            self.theEnsemble.mechanical_simulation_step(cool_factor=1, dt=.1)
            self.theTk.update()
            self.theEvaluator.evaluate_stress_tensors(self.theEnsemble.force_register, self.theEnsemble.sphereList, 0)

        stress_tensor_torque = self.theEvaluator.stress_tensor_unbalanced_torque

        self.assertTrue(
             stress_tensor_torque[0][1] == -stress_tensor_torque[1][0])


    def test_conservation_linear_momentum(self):

        self.theEnsemble.mechanical_simulation_step_calculate_forces()

        self.theEnsemble.record_total_particle_forces()
        self.theEvaluator.evaluate_stress_tensors(self.theEnsemble.force_register, self.theEnsemble.sphereList, 0)
        theHandler = EvaluationHandler()

        theHandler.record(self.theEnsemble, self.theEvaluator)

        self.theEnsemble.mechanical_simulation_step_calculate_acceleration(cool_factor=self.cool_factor, dt=self.dt)
        self.theEnsemble.mechanical_simulation_step_calculate_movement(dt=self.dt)
        self.theTk.update()

        old_linear_momentum = theHandler.total_mass * theHandler.v_mean

        for i in range(100):
            self.theEnsemble.mechanical_simulation_step(cool_factor=1, dt=self.dt)
            self.theTk.update()

        self.theEnsemble.mechanical_simulation_step_calculate_forces()

        self.theEnsemble.record_total_particle_forces()
        self.theEvaluator.evaluate_stress_tensors(self.theEnsemble.force_register, self.theEnsemble.sphereList, 0)
        theHandler = EvaluationHandler()

        theHandler.record(self.theEnsemble, self.theEvaluator)

        self.theEnsemble.mechanical_simulation_step_calculate_acceleration(cool_factor=self.cool_factor, dt=self.dt)
        self.theEnsemble.mechanical_simulation_step_calculate_movement(dt=self.dt)
        self.theTk.update()

        new_linear_momentum = theHandler.total_mass * theHandler.v_mean

        if(math.sqrt(new_linear_momentum[0]*new_linear_momentum[0]+\
                                        new_linear_momentum[1]*new_linear_momentum[1])>1e-5):

            self.assertTrue(abs(old_linear_momentum[0] - new_linear_momentum[0]) \
                            / math.sqrt(new_linear_momentum[0]*new_linear_momentum[0]+\
                                        new_linear_momentum[1]*new_linear_momentum[1]) < 1e-15)

            self.assertTrue(abs(old_linear_momentum[1] - new_linear_momentum[1]) \
                            / math.sqrt(new_linear_momentum[0]*new_linear_momentum[0]+\
                                        new_linear_momentum[1]*new_linear_momentum[1]) < 1e-15)

        else:
            self.assertTrue(abs(old_linear_momentum[0]-new_linear_momentum[0])<1e-15)
            self.assertTrue(abs(old_linear_momentum[1] - new_linear_momentum[1]) < 1e-15)

    def test_conservation_angular_momentum(self):

        self.theEnsemble.mechanical_simulation_step_calculate_forces()

        self.theEnsemble.record_total_particle_forces()
        self.theEvaluator.evaluate_stress_tensors(self.theEnsemble.force_register, self.theEnsemble.sphereList, 0)
        theHandler = EvaluationHandler()

        theHandler.record(self.theEnsemble, self.theEvaluator)

        self.theEnsemble.mechanical_simulation_step_calculate_acceleration(cool_factor=self.cool_factor, dt=self.dt)
        self.theEnsemble.mechanical_simulation_step_calculate_movement(dt=self.dt)
        self.theTk.update()

        old_angular_momentum = theHandler.angular_momentum_around_origin

        for i in range(100):
            self.theEnsemble.mechanical_simulation_step(cool_factor=1, dt=self.dt)
            self.theTk.update()

        self.theEnsemble.mechanical_simulation_step_calculate_forces()

        self.theEnsemble.record_total_particle_forces()
        self.theEvaluator.evaluate_stress_tensors(self.theEnsemble.force_register, self.theEnsemble.sphereList, 0)
        theHandler = EvaluationHandler()

        theHandler.record(self.theEnsemble, self.theEvaluator)

        self.theEnsemble.mechanical_simulation_step_calculate_acceleration(cool_factor=self.cool_factor, dt=self.dt)
        self.theEnsemble.mechanical_simulation_step_calculate_movement(dt=self.dt)
        self.theTk.update()

        new_angular_momentum = theHandler.angular_momentum_around_origin



        self.assertTrue(abs((old_angular_momentum - new_angular_momentum) / new_angular_momentum) < 1e-14)



if __name__ == '__main__':
    unittest.main()


from .doParticleShearSimulation import doParticleShearSimulation


def doParticleShearSimulationSeries(root_folder="",amplitudes=[0.2],mu=[0.01],cut_lines=5,Young_modulus_spheres=8000,
                    do_bulk_control=False,do_uncrosslinked_control=False,
                    do_non_frictional_control=False,N_repetitions=6,
                    N=150,packing_fraction=1.5,
                    density=50,
                    bimodal_factor=1.4,
                    relative_viscosity=0.1,relative_transversal_link_strength=1,
                    relative_frequency=0.025,
                    avoid_horizontal_angle_degree=15,
                    interface_reenforcement_central=1,interface_reenforcement_tangential=1,
                    keep_viscosity_coefficients_constant=False,
                    avoid_height_spanning_particles=True,cut_top_bottom=False,
                    doCutByTriangulation=True,
                    saveOutputImages=False,imageFileType = "jpg",
                    remove_link_fraction=0,
                    edge_fuzziness=0,central_repulsion_coefficient=1,
                    doDrawing=True,saveData=False,saveStressTensorData=True,plotStress=True,
                    pre_periods=2, periods=3, post_periods=1, cool_factor=0.5):

    for i in range(N_repetitions):
        for theAmplitude in amplitudes:


            for theMu in mu:
                # Control: Bulk (this is the same as the biomaterial call, but cut_lines=0
                if do_bulk_control:

                    doParticleShearSimulation(root_folder,theAmplitude=theAmplitude,
                        theMu=theMu,do_permanent_links=True,cut_lines=0,
                        Young_modulus_spheres=Young_modulus_spheres,
                        N=N,packing_fraction=packing_fraction,
                        density=density,
                        bimodal_factor=bimodal_factor,
                        relative_viscosity=relative_viscosity,
                        relative_transversal_link_strength=relative_transversal_link_strength,
                        relative_frequency=relative_frequency,
                        avoid_horizontal_angle_degree=avoid_horizontal_angle_degree,
                        interface_reenforcement_central=interface_reenforcement_central,
                        interface_reenforcement_tangential=interface_reenforcement_tangential,
                        keep_viscosity_coefficients_constant=keep_viscosity_coefficients_constant,
                        avoid_height_spanning_particles=avoid_height_spanning_particles,cut_top_bottom=cut_top_bottom,
                        saveOutputImages=saveOutputImages,imageFileType =imageFileType,
                        remove_link_fraction=remove_link_fraction,
                        edge_fuzziness=edge_fuzziness,
                        central_repulsion_coefficient=central_repulsion_coefficient,
                        doCutByTriangulation=doCutByTriangulation,
                        doDrawing=doDrawing,
                        saveData=saveData,
                        saveStressTensorData=saveStressTensorData,
                        plotStress=plotStress,
                        pre_periods=pre_periods,periods=periods,post_periods=post_periods,cool_factor=cool_factor)




                # Actual biomaterial: crosslinked but cut into particles
                doParticleShearSimulation(root_folder, theAmplitude=theAmplitude,
                    theMu=theMu, do_permanent_links=True, cut_lines=cut_lines,
                    Young_modulus_spheres=Young_modulus_spheres,
                    N=N, packing_fraction=packing_fraction,
                    density=density,
                    bimodal_factor=bimodal_factor,
                    relative_viscosity=relative_viscosity,
                    relative_transversal_link_strength=relative_transversal_link_strength,
                    relative_frequency=relative_frequency,
                    avoid_horizontal_angle_degree=avoid_horizontal_angle_degree,
                    interface_reenforcement_central=interface_reenforcement_central,
                    interface_reenforcement_tangential=interface_reenforcement_tangential,
                    keep_viscosity_coefficients_constant=keep_viscosity_coefficients_constant,
                    avoid_height_spanning_particles=avoid_height_spanning_particles,
                    cut_top_bottom=cut_top_bottom,
                    saveOutputImages=saveOutputImages,imageFileType =imageFileType,
                    remove_link_fraction=remove_link_fraction,
                    edge_fuzziness=edge_fuzziness,
                    central_repulsion_coefficient=central_repulsion_coefficient,
                    doCutByTriangulation=doCutByTriangulation,
                    doDrawing=doDrawing,
                    saveData=saveData,
                    saveStressTensorData=saveStressTensorData,
                    plotStress=plotStress,
                    pre_periods=pre_periods, periods=periods, post_periods=post_periods,
                    cool_factor=cool_factor)

                # Control: uncrosslinked spheres only
                if do_uncrosslinked_control:
                    doParticleShearSimulation(root_folder, theAmplitude=theAmplitude,
                        theMu=theMu, do_permanent_links=False, cut_lines=0,
                        Young_modulus_spheres=Young_modulus_spheres,
                        N=N, packing_fraction=packing_fraction,
                        density=density,
                        bimodal_factor=bimodal_factor,
                        relative_viscosity=relative_viscosity,
                        relative_transversal_link_strength=relative_transversal_link_strength,
                        relative_frequency=relative_frequency,
                        avoid_horizontal_angle_degree=avoid_horizontal_angle_degree,
                        interface_reenforcement_central=interface_reenforcement_central,
                        interface_reenforcement_tangential=interface_reenforcement_tangential,
                        keep_viscosity_coefficients_constant=keep_viscosity_coefficients_constant,
                        avoid_height_spanning_particles=avoid_height_spanning_particles,
                        cut_top_bottom=cut_top_bottom,
                        saveOutputImages=saveOutputImages,imageFileType =imageFileType,
                        remove_link_fraction=remove_link_fraction,
                        edge_fuzziness=edge_fuzziness,
                        central_repulsion_coefficient=central_repulsion_coefficient,
                        doCutByTriangulation=doCutByTriangulation,
                        doDrawing=doDrawing,
                        saveData=saveData,
                        saveStressTensorData=saveStressTensorData,
                        plotStress=plotStress,
                        pre_periods=pre_periods, periods=periods, post_periods=post_periods,
                        cool_factor=cool_factor)


                # Control: non-frictional
                if do_non_frictional_control:
                    doParticleShearSimulation(root_folder, theAmplitude=theAmplitude,
                        theMu=0, do_permanent_links=True, cut_lines=cut_lines,
                        Young_modulus_spheres=Young_modulus_spheres,
                        N=N, packing_fraction=packing_fraction,
                        density=density,
                        bimodal_factor=bimodal_factor,
                        relative_viscosity=relative_viscosity,
                        relative_transversal_link_strength=relative_transversal_link_strength,
                        relative_frequency=relative_frequency,
                        avoid_horizontal_angle_degree=avoid_horizontal_angle_degree,
                        interface_reenforcement_central=interface_reenforcement_central,
                        interface_reenforcement_tangential=interface_reenforcement_tangential,
                        keep_viscosity_coefficients_constant=keep_viscosity_coefficients_constant,
                        avoid_height_spanning_particles=avoid_height_spanning_particles,
                        cut_top_bottom=cut_top_bottom,
                        saveOutputImages=saveOutputImages,imageFileType =imageFileType,
                        remove_link_fraction=remove_link_fraction,
                        edge_fuzziness=edge_fuzziness,
                        central_repulsion_coefficient=central_repulsion_coefficient,
                        doCutByTriangulation=doCutByTriangulation,
                        doDrawing=doDrawing,
                        saveData=saveData,
                        saveStressTensorData=saveStressTensorData,
                        plotStress=plotStress,
                        pre_periods=pre_periods, periods=periods, post_periods=post_periods,
                        cool_factor=cool_factor)





from tkinter import *
import random
import math
from .CanvasPointsBasicElasticityLeesEdwards import CanvasPointsBasicElasticityLeesEdwards
from .Force_register import Force_register

class CanvasPointsFrictionElasticityLeesEdwards(CanvasPointsBasicElasticityLeesEdwards):
    """Canvas for placing `particleShear.CircleFrictionElasticity` objects (referred to as spheres).
          This class maintains a list of objects of type
          `particleShear.CircleFrictionElasticity` or derived and adds functions for handling friction and associated
          non-central forces"""

    def __init__(self, size_x, size_y, theCanvas=FALSE, doDrawing=FALSE,
                 k=1, nu=0.01,m=1,k_t=1,nu_t=0.01,mu=0.1):
        """Initialize self

                        - **parameters**\n
                            `size_x` Width of area to be used in pixels = micrometers for the simulation\n
                            `size_y` Height of area to be used in pixels = micrometers for the simulation\n
                            `theCanvas` possibility to transmit a `tkinter` `Canvas` object for graphical output\n
                            `doDrawing` Flag to indicate whether graphical output should be produced or not\n
                            `k` Central spring constant in (mg/s^2)/m of depth;
                            so F=-k*x*L with L the depth of the stack (the simulation is 2D) and x the compression\n
                            `nu` Central viscosity force constant in (mg/s)/m of depth. This is to have F=nu*v*L, L again the depth\n
                            `m` is the mass per unit of depth in mg/m of depth\n
                            `k_t` Tangential spring constant in (mg/s^2)/m of depth;
                            relevant for locked or permanent interfaces but not for slipping interfaces\n
                            `nu_t` Central viscosity force constant in (mg/s)/m of depth.
                             Relevant for frictionally locked and permanent interfaces but not slipping ones\n
                            `mu` Friction coefficient, describes the maximum interface force for non-permanent interfaces by
                            F_tangential_max = F_central*mu"""



        super(CanvasPointsFrictionElasticityLeesEdwards, self).__init__( size_x, size_y, theCanvas, doDrawing,
                                                          k, nu, m)
        self.k_t=k_t
        """`k_t` Tangential spring constant in (mg/s^2)/m of depth;"""
        self.nu_t=nu_t
        """`nu_t` Central viscosity force constant in (mg/s)/m of depth."""
        self.mu=mu
        """`mu` Friction coefficient"""

    def particle_info(self):
        return "CanvaPointsFrictionElasticityLeesEdwards: Lees Edwards boundary conditions and stick-and-slip type friction"


    def tangential_force(self):
        """Let all spheres calculate the total tangental force acting on them.

         This will check for contact and so is more time
         consuming than `particleShear.CanvasPointsFrictionElasticityLeesEdwards.tangential_force_from_neighbors`.
         For this function to work, the spheres must be of type `particleShear.CircleFrictionElasticity`or a
         subclass thereof, or otherwise possess a method `tangential_force`\n
         This method is defined in class `particleShear.CanvasPointsFrictionElasticityLeesEdwards`"""

        for sphereIndex in range(len(self.sphereList)):
            for sphereIndex2 in range(len(self.sphereList)):
                if sphereIndex != sphereIndex2:
                    self.sphereList[sphereIndex].tangential_force(self.sphereList[sphereIndex2],
                                                                  self.nu_t, self.mu, self.k, self.k_t)

    def tangential_force_from_neighbors(self):
        """Let all spheres calculate the total tangential force acting on from their known neighbors.

        This is quicker than `particleShear.CanvasPointsFrictionElasticityLeesEdwards.tangential_force`
        but necessitates the neighbor relations to be already established via
        `particleShear.CanvasPointsNeighbors.test_neighbor_relation`.
        For this function to work, the spheres must be of type `particleShear.CircleFrictionElasticity`or a
        subclass thereof, or otherwise possess a method `tangential_force_from_neighbors`.\n
        This method is defined in class `particleShear.CanvasPointsFrictionElasticityLeesEdwards`"""
        for sphereIndex in range(len(self.sphereList)):
            self.sphereList[sphereIndex].tangential_force_from_neighbors(self.nu_t,self.mu,self.k,self.k_t)

    def do_rotational_acceleration(self, dt):
        """Have the spheres calculate their rotational acceleration from the tangential forces.

        For this function to work,
        the spheres must be of type `particleShear.CircleFrictionElasticity` or subclass thereof, or alternatively,
        possess a method `do_rotational_acceleration`\n
        This method is defined in class `particleShear.CanvasPointsFrictionElasticityLeesEdwards`"""
        for theSphere in self.sphereList:
            theSphere.do_rotational_acceleration(dt)


    def mechanical_simulation_step_calculate_forces(self,dt=1):
        super(CanvasPointsFrictionElasticityLeesEdwards,self).mechanical_simulation_step_calculate_forces()
        self.tangential_force_from_neighbors()

    def record_total_particle_forces(self):

        super(CanvasPointsFrictionElasticityLeesEdwards,self).record_total_particle_forces()

        for theSphere in self.sphereList:
            moment = theSphere.torque
            self.force_register.record_unbalanced_moment(theSphere,moment)

    def mechanical_simulation_step_calculate_movement(self,dt=1):

        self.move(dt)
        self.setShear(self.shear + self.shear_rate * dt)
        self.setShearRate(self.shear_rate)
        self.boundary_conditions()
        self.t = self.t + dt


    def mechanical_simulation_step_calculate_acceleration(self,cool_factor=0.97,dt=1):
        self.do_linear_acceleration(dt)
        self.do_rotational_acceleration(dt)
        self.cool(cool_factor)



























from particleShearSimulation import Simulation_interlocking_rheology


def doParticleShearSimulation(root_folder="",theAmplitude=0.2,theMu=0.01,do_permanent_links=True,
                    cut_lines=5,Young_modulus_spheres=8000,
                    N=150,packing_fraction=1.5,density=50,bimodal_factor=1.4,
                    relative_viscosity=0.1,relative_transversal_link_strength=1,relative_frequency=0.025,
                    avoid_horizontal_angle_degree=15,interface_reenforcement_central=1,interface_reenforcement_tangential=1,
                    keep_viscosity_coefficients_constant=False,avoid_height_spanning_particles=True,cut_top_bottom=False,
                    saveOutputImages=False,imageFileType ="jpg",
                    remove_link_fraction=0,edge_fuzziness=0,
                    central_repulsion_coefficient=1,
                    doCutByTriangulation=True,doDrawing=True,saveData=False,
                    saveStressTensorData=True,
                    plotStress=True,relative_y_scale_force = 1e8,
                    pre_periods=2,periods=3,post_periods=1,cool_factor=0.5):


    #if theAmplitude < 0.005:
    #    relative_y_scale_force = relative_y_scale_force * theAmplitude / 0.005



    theSimulation = Simulation_interlocking_rheology(
        root_folder, do_permanent_links=do_permanent_links, cut_lines=cut_lines, N=N,
        packing_fraction=packing_fraction,
        density=density,
        bimodal_factor=bimodal_factor,
        amplitude=theAmplitude, Young_modulus_spheres=Young_modulus_spheres, mu=theMu,
        relative_y_scale_force=relative_y_scale_force, relative_frequency=relative_frequency, relative_viscosity=relative_viscosity,
        relative_transversal_link_strength=relative_transversal_link_strength,
        central_repulsion_coefficient=central_repulsion_coefficient,
        avoid_horizontal_angle_degree=avoid_horizontal_angle_degree,
        interface_reenforcement_central=interface_reenforcement_central,
        interface_reenforcement_tangential=interface_reenforcement_tangential,
        keep_viscosity_coefficients_constant=keep_viscosity_coefficients_constant,
        avoid_height_spanning_particles=avoid_height_spanning_particles,
        cut_top_bottom=cut_top_bottom,
        doCutByTriangulation=doCutByTriangulation,
        remove_link_fraction=remove_link_fraction,
        edge_fuzziness=edge_fuzziness,
        doDrawing=doDrawing,
        saveData=saveData
    )



    theSimulation.function_call = theSimulation.function_call + \
                                  "\n\ttheSimulation.baseline_pre_periods = " + str(pre_periods) +\
                                  "\n\ttheSimulation.baseline_post_periods = " + str(post_periods)+ \
                                  "\n\ttheSimulation.saveOutputImages = " + str(saveOutputImages)+ \
                                  "\n\ttheSimulation.imageFileType = " + str(imageFileType) + \
                                  "\n\ttheSimulation.runSimulation(periods="+str(periods)+", cool_factor="+\
                                  str(cool_factor)+")"

    theSimulation.baseline_pre_periods = pre_periods
    theSimulation.baseline_post_periods = post_periods

    theSimulation.saveOutputImages = saveOutputImages
    theSimulation.imageFileType = imageFileType


    theSimulation.saveStressTensorData=saveStressTensorData

    return(theSimulation.runSimulation(periods=periods, cool_factor=cool_factor,plotStress=plotStress))






"""Provide geometrical simulation elements and canvas areas to place them"""

__all__ = ["neighbor_relation_linkable",
           "CanvasPointsLinkable","SphereLinkable","SphereLinkableAdjustableInterfaceStrength",
"EnsembleLinkable","EnsembleCompactParticles",
"EnsembleCompactParticlesAdjustableInterfaceStrength",
           "EnsembleCompactParticlesAdjustableInterfaceStrengthFromModelParameters",
           "EnsembleCompactParticlesFromModelParameters",
           "SphereLinkable","SphereLinkableAdjustableInterfaceStrength",
           "elastic_force_law_tensile","elastic_force_law_tensile_exponential",
           "TensileConfiguration"]

# Basic tools
from .neighbor_relation_linkable import neighbor_relation_linkable


# Geometrical elements
from .CanvasPointsLinkable import CanvasPointsLinkable

# Particular ensemble definitions
from .EnsembleLinkable import EnsembleLinkable
from .EnsembleCompactParticles import EnsembleCompactParticles
from .EnsembleCompactParticlesAdjustableInterfaceStrength import EnsembleCompactParticlesAdjustableInterfaceStrength
from .EnsembleCompactParticlesAdjustableInterfaceStrength \
    import EnsembleCompactParticlesAdjustableInterfaceStrengthFromModelParameters
from .EnsembleCompactParticles import EnsembleCompactParticlesFromModelParameters





# Sphere definitions
from .SphereLinkable import SphereLinkable
from .SphereLinkableAdjustableInterfaceStrength import SphereLinkableAdjustableInterfaceStrength


# Alternative tensile laws
from .SphereLinkable import TensileConfiguration
from .SphereLinkable import elastic_force_law_tensile
from .SphereLinkable import elastic_force_law_tensile_exponential











from tkinter import *
from particleShearBase import *
from particleShearObjects import *
from particleShearLinkableObjects import *
from .EvaluationHandlerPlotter import EvaluationHandlerPlotter
from .EvaluationHandler import  EvaluationHandler
import math
import io
from PIL import Image



class OscillatoryShearExperiment():
    # The plotStress variable serves as a switch to enable/disable output plotting, while leaving everything else the same
    def __init__(self, theEnsemble,dt,f,amplitude,output_file,dt_max=0,TkSimulation=False,force_scale=1,theTkOutput=False,
                 imageOutputFolder=False,imageBaseFileName=False,imageFileType = "jpg",plotStress=True,
                 saveStressTensorData=True):
        self.plotStress=plotStress
        print("OscillatoryShearExperiment: Plot stress graph:",self.plotStress)
        self.theEnsemble=theEnsemble
        self.amplitude=amplitude
        self.dt = dt
        self.f = f
        self.theTk = theTkOutput
        # The idea here is to be able to handle an external Tk so that the rest of the program can do something with the
        # graphical output
        self.external_output_Tk=False
        if self.theTk:
            self.external_output_Tk = True
        self.width=1000
        self.height=500
        self.outputCanvas = False
        self.output_file = False
        self.dt_max = 10*dt
        self.theTkSimulation=TkSimulation
        if dt_max>dt:
            self.dt_max = dt_max

        self.plotter = False
        self.theEvaluator = False
        self.baseline_pre_periods = 0 # Acquisition of stress data with 0
            # excitation amplitude to have a baseline before oscillatory shear
        self.baseline_post_periods = 0 # Acquisition of stress data with 0
        # excitation amplitude to have a baseline before oscillatory shear
        self.periods=2
        self.force_scale=force_scale
        self.output_file=output_file

        self.final_cool_factor=1

        self.do_preequilibration_with_fixed_boundaries=True


        self.imageOutputFolder=imageOutputFolder

        self.imageBaseFileName = imageBaseFileName

        self.imageFileType = imageFileType # can be either jpg or ps

        self.saveStressTensorData=saveStressTensorData





    def free_pre_equilibration(self):


        print("Free pre-equilibration, dt_max",self.dt_max, "dt = ",self.dt)
        self.theEnsemble.applyingShear = False
        dt=self.dt_max
        N = 10
        cool_factor=0.5
        if self.theEnsemble.mu > 0:
            N=int(20 * (2 - math.log10(self.theEnsemble.mu)))
        while dt>self.dt:
            print("dt=", dt)
            self.theEnsemble.mechanical_relaxation(dt=dt, theTk=self.theTkSimulation,
                    N=N, cool_factor=cool_factor)
            dt=dt/2
            cool_factor=(1+cool_factor)/2
            self.theEnsemble.correct_linear_drift()
        dt = self.dt
        print("dt=", dt)
        self.theEnsemble.mechanical_relaxation(dt=dt, theTk=self.theTkSimulation,
                    N=N, cool_factor=1)


    def pre_equilibrate(self,  N=25,cool_factor=0.99):

        print("Pre-equilibration under experimental conditions, Steps per dt:",N," max dt",self.dt_max," final dt",self.dt)

        self.theEnsemble.applyingShear = True
        dt = self.dt_max
        cf=0.5

        while dt > self.dt:
            print("dt=", dt)
            self.theEnsemble.mechanical_relaxation(dt=dt, theTk=self.theTkSimulation,
                                                   N=int(N/2), cool_factor=cf)
            dt = dt / 2
            cf = (1 + cf) / 2

        dt = self.dt
        print("  Setting final dt=", dt)
        self.theEnsemble.mechanical_relaxation(dt=dt, theTk=self.theTkSimulation,
                                               N=N*2, cool_factor=0.98)


        for i in range(N):
            self.theEnsemble.shear_rate = 0
            self.theEnsemble.mechanical_simulation_step(cool_factor, self.dt)
            self.theEnsemble.reset_force()
            if self.theEnsemble.doDrawing:
                self.theTkSimulation.update()


    def oscillatory_shear_experiment(self, cool_factor=1,N_pre_equilibration_fixed_boundaries=25):


        self.theEvaluator = StressTensorEvaluation(self.theEnsemble.size_x, self.theEnsemble.size_y,
                                              self.theEnsemble.theCanvas)

        self.final_cool_factor = cool_factor

        self.write_experiment_description_information_to_file()

        if(self.do_preequilibration_with_fixed_boundaries):
            self.pre_equilibrate(cool_factor=cool_factor,N=N_pre_equilibration_fixed_boundaries)

        self.theEnsemble.t=0

        self.theEnsemble.applyingShear = True

        if self.plotStress and not self.theTk:

            self.theTk=Tk()


        if self.plotStress:
            self.outputCanvas = Canvas(self.theTk, width=self.width, height=self.height)

            self.outputCanvas.pack()



        self.theEnsemble.t=0



        Time_required = (self.periods+self.baseline_pre_periods+self.baseline_post_periods) / self.f

        x_scale = self.width/Time_required



        self.plotter=EvaluationHandlerPlotter(self.theTk, self.outputCanvas, x_scale=x_scale,
                                 y_scale_force=self.force_scale,
                                 y_scale_strain=30 / self.amplitude, r=2, offset_x=0,
                                 offset_y=self.height / 2,plotStress=self.plotStress)



        N=int(Time_required/self.dt)

        print("Start application shear protocol, total N=",N," steps")

        # plan to output 200 images in total

        image_N=1

        if N>200:
            image_N=int(N/200)

        self.theEnsemble.t=0

        imageCounter=image_N

        if self.imageOutputFolder:
            print("Saving image every ",image_N," steps")

        for i in range(N):
            if i>=(self.baseline_pre_periods/self.f/self.dt) and i<=((self.periods+self.baseline_pre_periods)/self.f/self.dt):
                self.theEnsemble.setShearRate(
                    self.amplitude * self.out_of_phase_signal(self.theEnsemble.t)* self.f * 2 * math.pi)



            else:
                self.theEnsemble.setShearRate(0)

            # Calculate forces, this registers individual forces in the force register

            self.theEnsemble.mechanical_simulation_step_calculate_forces()

            # report net particles forces and moments in the force register
            self.theEnsemble.record_total_particle_forces()


            self.plotter.record(self.theEnsemble,self.theEvaluator)



            # Do linear and rotational acceleration
            self.theEnsemble.mechanical_simulation_step_calculate_acceleration(cool_factor=cool_factor, dt=self.dt)
            # Do linear and rotation movememnt
            self.theEnsemble.mechanical_simulation_step_calculate_movement(dt=self.dt)
            # reset forces for next round
            self.theEnsemble.reset_force()

            if self.theEnsemble.doDrawing:
                self.theTkSimulation.update()

            if imageCounter>=image_N:

                self.save_canvas_image(i)
                imageCounter=0

            imageCounter=imageCounter+1


        print("overall", self.plotter.overall_stress_tensor[len(self.plotter.shear_stress) - 1])


        print("overall", self.plotter.overall_stress_tensor[0])

        G=self.write_output_information_to_file()



        if not self.external_output_Tk:
            if self.theTk:
                self.theTk.destroy()
                self.theTk = False
                self.outputCanvas = False

        return(G)


    def  save_canvas_image(self,index):
        if not self.imageOutputFolder or not self.theEnsemble.doDrawing or not self.theEnsemble.theCanvas:
            return

        filename=self.imageOutputFolder+"/"+self.imageBaseFileName
        if index < 1e6:
            filename=filename+"0"
        if index < 1e5:
            filename=filename+"0"
        if index < 1e4:
            filename=filename+"0"
        if index < 1e3:
            filename=filename+"0"
        if index < 1e2:
            filename = filename + "0"
        if index < 1e1:
            filename = filename + "0"

        file = filename + str(index)

        if self.imageFileType=="jpg":
            file=file+ ".jpg"
        else:
            file=file+".ps"

        if self.imageFileType=="jpg":
            ps=self.theEnsemble.theCanvas.postscript(colormode="color")
            img = Image.open(io.BytesIO(ps.encode('utf-8')))
            img.save(file)
        else:
            self.theEnsemble.theCanvas.postscript(file=file,colormode="color")









    def in_phase_signal(self,t):
        return math.sin((t * self.f - self.baseline_pre_periods) * math.pi * 2)

    def out_of_phase_signal(self,t):
        return math.cos((t * self.f - self.baseline_pre_periods) * math.pi * 2)


    # Returns a vector of four elements: the demodulated stress (in phase), the demodulated stress (90 degress),
    # G' and G''
    def demodulation(self,t,shear_stress):
        t = t
        in_phase = []
        out_of_phase = []
        sum_in_phase = 0
        sum_out_of_phase = 0
        stress_in_phase = 0
        stress_out_of_phase = 0
        stress_in_phase_linear_drift_corrected = 0
        stress_out_of_phase_linear_drift_corrected = 0

        # If we have more than one period available, do not use the first one for evaluation as there are transitory
        # phenomena at the beginning
        equilibration_periods = 0
        if self.periods >= 2:
            equilibration_periods = 1

        lastIndex = len(shear_stress) - 1

        lastStressForDelta = shear_stress[lastIndex]
        firstStressForDelta = shear_stress[0]

        firstTForDelta = t[0]
        lastTForDelta = t[lastIndex]

        if self.baseline_pre_periods > 0:
            sum_stress = 0
            sum_t = 0
            N = 0
            for i in range(len(t)):
                if t[i] < (self.baseline_pre_periods / self.f) and t[i] >= (2 * self.baseline_pre_periods / self.f / 3):
                    N = N + 1
                    sum_stress = sum_stress + shear_stress[i]
                    sum_t = sum_t + t[i]
            firstStressForDelta = sum_stress / N
            firstTForDelta = sum_t / N

        if self.baseline_post_periods > 0:
            sum_stress = 0
            sum_t = 0
            N = 0
            for i in range(len(t)):
                if t[i] > (self.baseline_pre_periods + self.periods) / self.f and t[i] <= \
                        (self.baseline_pre_periods + self.periods + self.baseline_post_periods / 3) / self.f:
                    N = N + 1
                    sum_stress = sum_stress + shear_stress[i]
                    sum_t = sum_t + t[i]
            lastStressForDelta = sum_stress / N
            lastTForDelta = sum_t / N

        linear_drift_coefficient = (lastStressForDelta - firstStressForDelta) / (firstTForDelta - lastTForDelta)



        for i in range(len(t)):
            ip = 0
            op = 0
            if t[i] >= ((self.baseline_pre_periods + equilibration_periods) / self.f) and t[i] <= (
                    (self.baseline_pre_periods + self.periods) / self.f):
                ip = self.in_phase_signal(t[i])
                op = self.out_of_phase_signal(t[i])

            in_phase.append(ip)
            out_of_phase.append(op)

            sum_in_phase = sum_in_phase + ip * ip
            sum_out_of_phase = sum_out_of_phase + op * op

            stress_linear_drift_corrected = shear_stress[i] - (
                    t[i] - self.baseline_pre_periods / self.f) * linear_drift_coefficient

            stress_in_phase = stress_in_phase + shear_stress[i] * ip
            stress_out_of_phase = stress_out_of_phase + shear_stress[i] * op

            stress_in_phase_linear_drift_corrected = stress_in_phase_linear_drift_corrected + stress_linear_drift_corrected * ip
            stress_out_of_phase_linear_drift_corrected = stress_out_of_phase_linear_drift_corrected + stress_linear_drift_corrected * op


        return [stress_in_phase_linear_drift_corrected / sum_in_phase,
                stress_out_of_phase_linear_drift_corrected / sum_out_of_phase,
                stress_in_phase_linear_drift_corrected / sum_in_phase / self.amplitude,
                stress_out_of_phase_linear_drift_corrected / sum_out_of_phase / self.amplitude]


    def evaluateStressAndG(self,useExternalForce=False):
        if not useExternalForce:
            return self.demodulation(self.plotter.t,self.plotter.shear_stress)
        else:
            return self.demodulation(self.plotter.t, self.plotter.force)



    def write_experiment_description_information_to_file(self):

        if not self.output_file:
            print("Not saving experiment description to file ( no data file configured )")
            return
        file=self.output_file

        print("Saving simulation description to output file")

        file.write("Oscillatory shear experiment\n\n")


        if hasattr(self.theEnsemble,"model"):
            file.write("Model parameters\n")
            file.write("Width of simulation area size_x =")
            file.write(str(self.theEnsemble.model.size_x))
            file.write("micrometers\n")
            file.write("Height of simulation area size_y =")
            file.write(str(self.theEnsemble.model.size_y))
            file.write("micrometers\n")
            file.write("Nominal packing fraction = ")
            file.write(str(self.theEnsemble.model.packing_fraction))
            file.write("\n")
            file.write("Young modulus spheres = ")
            file.write(str(self.theEnsemble.model.Young_modulus_spheres))
            file.write("Pa\n")
            file.write("Relative density spheres = ")
            file.write(str(self.theEnsemble.model.density))
            file.write("kg/m^3\n")
            file.write("Characteristic time constant = ")
            file.write(str(self.theEnsemble.model.time_constant))
            file.write("s\n")
            file.write("Characteristic spring constant = ")
            file.write(str(self.theEnsemble.model.k))
            file.write("mg/s^2\n")
            file.write("Characteristic viscosity  = ")
            file.write(str(self.theEnsemble.model.k*self.theEnsemble.model.time_constant))
            file.write("mg/s\n\n")

        file.write("Ensemble parameters\n")

        file.write("Particle mass m = ")
        file.write(str(self.theEnsemble.m))
        file.write("mg\n")

        file.write("Central spring constant k = ")
        file.write(str(self.theEnsemble.k))
        file.write("mg*s^(-2)\n")

        file.write("Central viscosity constant nu = ")
        file.write(str(self.theEnsemble.nu))
        file.write("mg*s^(-1)\n")

        file.write("Lateral spring constant k_t = ")
        file.write(str(self.theEnsemble.k_t))
        file.write("mg*s^(-2)\n")

        file.write("Lateral viscosity constant nu_t = ")
        file.write(str(self.theEnsemble.nu_t))
        file.write("mg*s^(-1)\n")

        file.write("Friction coefficient mu = ")
        file.write(str(self.theEnsemble.mu))
        file.write("\n")

        file.write("Sample width = ")
        file.write(str(self.theEnsemble.size_x))
        file.write("micrometers\n")

        file.write("Sample height = Gap height = ")
        file.write(str(self.theEnsemble.size_y))
        file.write("micrometers\n")

        file.write("Particle number = ")
        file.write(str(self.theEnsemble.N))
        file.write("\n")

        file.write("Particle information\n\t")
        file.write(str(self.theEnsemble.particle_info()))
        file.write("\n")


        if hasattr(self.theEnsemble,"k_corner_torque"):
            file.write("Angular torque spring constant, per m of depth: k_corner_torque = ")
            file.write(str(self.theEnsemble.k_corner_torque))
            file.write("mg/m*micrometers^2/s^-2\n")

        if hasattr(self.theEnsemble, "nu_corner_torque"):
            file.write("Angular torque viscosity constant, per m of depth: nu_corner_torque = ")
            file.write(str(self.theEnsemble.nu_corner_torque))
            file.write("mg/m*micrometers^2/s^-1\n")


        if hasattr(self.theEnsemble,"central_repulsion_coefficient"):
            file.write("Nonlinear enhancement for repulsion near sphere center = ")
            file.write(str(self.theEnsemble.central_repulsion_coefficient))
            file.write("[-], 0=no enhancement, 1=maximum enhancement; see CircleBasicElasticity.get_elastic_force\n")


        if hasattr(self.theEnsemble,"permanent_ratio_central"):
            file.write("Adjustment coefficient for central force in permanent links = ")
            file.write(str(self.theEnsemble.permanent_ratio_central))
            file.write("[-], 0=no central force, 1=as in free interfaces; >1 re-enforcement of permanent over transient links\n")
            file.write("\tk_permanent=")
            file.write(str(self.theEnsemble.k * self.theEnsemble.permanent_ratio_central))
            file.write("mg*s^(-2)\n")

        if hasattr(self.theEnsemble,"permanent_ratio_tangential"):
            file.write("Adjustment coefficient for tangential force in permanent links = ")
            file.write(str(self.theEnsemble.permanent_ratio_tangential))
            file.write("[-], 0=no tangential force, 1=as in free interfaces; >1 re-enforcement of permanent over transient links\n")
            file.write("\tk_t_permanent=")
            file.write(str(self.theEnsemble.k_t * self.theEnsemble.permanent_ratio_tangential))
            file.write("mg*s^(-2)\n")


        if hasattr(self.theEnsemble,"keep_viscosity_coefficients_constant") and hasattr(self.theEnsemble,"permanent_ratio_tangential") \
                and hasattr(self.theEnsemble,"permanent_ratio_central"):
            if self.theEnsemble.keep_viscosity_coefficients_constant:
                file.write("Viscosity adjustment for permanent links:\n"
                           "\tNo adjustment: Impose constant viscosity regardless of the permanent link adjustment ")
                file.write("\n\tnu_permanent="+str(self.theEnsemble.nu))
                file.write("\n\tnu_t_permanent=" + str(self.theEnsemble.nu_t))
            else:
                file.write("Viscosity adjustment for permanent links:\n"
                           "\tAdjustment: Adjust viscosity (central and tangential) by same factor "+
                            "as the respective spring constant (k and k_t)")
                file.write("\n\tnu_permanent=" + str(self.theEnsemble.nu*self.theEnsemble.permanent_ratio_central))
                file.write("\n\tnu_t_permanent=" + str(self.theEnsemble.nu_t*self.theEnsemble.permanent_ratio_tangential))

        if not hasattr(self.theEnsemble,"avoid_height_spanning_particles"):
            file.write("\navoid_height_spanning_particles: Full height spanning particles allowed")
        else:
            if not self.theEnsemble.avoid_height_spanning_particles:
                file.write("\navoid_height_spanning_particles: Full height spanning particles allowed")
            else:
                file.write("\navoid_height_spanning_particles: Ensembles with full height spanning particles discarded")

        if hasattr(self.theEnsemble, "removed_fraction"):
            file.write("\nActual fraction of randomly removed non-essential permanent bonds = ")
            file.write(str(self.theEnsemble.removed_fraction))
            file.write(
                "[-]\n")

        if hasattr(self.theEnsemble, "edge_fuzziness"):
            file.write("\nEdge fuzziness for creating interlocking = ")
            file.write(str(self.theEnsemble.edge_fuzziness))
            file.write(
                "[-]\n")









        file.write("\n\nRheology experiment information\n")
        file.write("Simulated time = ")
        file.write(str(self.periods / self.f))
        file.write("s\n")

        file.write("Time step = ")
        file.write(str(self.dt))
        file.write("s\n")

        file.write("Frequency = ")
        file.write(str(self.f))
        file.write("Hz\n")

        file.write("Strain amplitude = ")
        file.write(str(self.amplitude*100))
        file.write("%\n")

        file.write("Damping factor = ")
        file.write(str(self.final_cool_factor))
        file.write(" [1 = no damping]\n")

        file.write("Pre-measurement periods = ")
        file.write(str(self.baseline_pre_periods))
        file.write(" [-]\n")

        file.write("Oscillatory shear periods = ")
        file.write(str(self.periods))
        file.write(" [-]\n")

        file.write("Post-measurement periods = ")
        file.write(str(self.baseline_pre_periods))
        file.write(" [-]\n")

        if hasattr(self.theEnsemble,"shear_rate_to_rotation_rate_coefficient"):
            file.write("Sllod equations: Anticipated adjustment of rotation for shear, with coefficient")
            file.write(str(self.theEnsemble.shear_rate_to_rotation_rate_coefficient))
            file.write(" [-]\n")

    def write_output_information_to_file(self):

        G=self.evaluateStressAndG(useExternalForce=False)
        G_external=self.evaluateStressAndG(useExternalForce=True)

        print("G'=")
        print(str(G[2]))
        print("Pa\n")
        print("G''=")
        print(str(G[3]))
        print("Pa\n")
        print("G'(by surface force)=")
        print(str(G_external[2]))
        print("Pa\n")
        print("G''(by surface force)=")
        print(str(G_external[3]))
        print("Pa\n")

        if not self.output_file:
            print("Not saving output to file (no data file configured)")
            return [G_external[2],G_external[3]]

        print("Saving simulation data to output file")
        file=self.output_file
        file.write("\n\nOutput data\nSummary data\n")
        file.write("In phase stress = ")
        file.write(str(G[0]))
        file.write("Pa\n")
        file.write("Out of phase stress = ")
        file.write(str(G[1]))
        file.write("Pa\n")
        file.write("Stress = ")
        file.write(str(math.sqrt(G[0]*G[0]+G[1]*G[1])))
        file.write("Pa\n")
        file.write("G'=")
        file.write(str(G[2]))
        file.write("Pa\n")
        file.write("G''=")
        file.write(str(G[3]))
        file.write("Pa\n")
        file.write("G'(by surface force)=")
        file.write(str(G_external[2]))
        file.write("Pa\n")
        file.write("G''(by surface force)=")
        file.write(str(G_external[3]))
        file.write("Pa\n")




        file.write("\n\nDetailed stress tensor data\n")

        if(not self.saveStressTensorData):

            print("Saving summary data only to file")

            file.write("\tNot saving detailed stress tensor data. If you wish to save this data, pass \n"+
                       "\t  saveStressTensorData=True")

        else:
            print("Saving stress tensor data to file")
            file.write("index\tt\tstrain\tstrain_rate\tstress_measured_at_surface\tshear_stress_internal_stress-tensor")
            file.write("\tstress_tensor_Love_Weber_00")
            file.write("\tstress_tensor_Love_Weber_01")
            file.write("\tstress_tensor_Love_Weber_10")
            file.write("\tstress_tensor_Love_Weber_11")
            file.write("\tstress_tensor_peculiar_acceleration_otsuki_00")
            file.write("\tstress_tensor_peculiar_acceleration_otsuki_01")
            file.write("\tstress_tensor_peculiar_acceleration_otsuki_10")
            file.write("\tstress_tensor_peculiar_acceleration_otsuki_11")
            file.write("\tstress_tensor_from_external_forces_00")
            file.write("\tstress_tensor_from_external_forces_01")
            file.write("\tstress_tensor_from_external_forces_10")
            file.write("\tstress_tensor_from_external_forces_11")
            file.write("\tstress_tensor_linear_acceleration_00")
            file.write("\tstress_tensor_linear_acceleration_01")
            file.write("\tstress_tensor_linear_acceleration_10")
            file.write("\tstress_tensor_linear_acceleration_11")
            file.write("\tstress_tensor_unbalanced_linear_forces_00")
            file.write("\tstress_tensor_unbalanced_linear_forces_01")
            file.write("\tstress_tensor_unbalanced_linear_forces_10")
            file.write("\tstress_tensor_unbalanced_linear_forces_11")
            file.write("\tstress_tensor_tangential_torque_00")
            file.write("\tstress_tensor_tangential_torque_01")
            file.write("\tstress_tensor_tangential_torque_10")
            file.write("\tstress_tensor_tangential_torque_11")
            file.write("\tstress_tensor_internal_tangential_torque_00")
            file.write("\tstress_tensor_internal_tangential_torque_01")
            file.write("\tstress_tensor_internal_tangential_torque_10")
            file.write("\tstress_tensor_internal_tangential_torque_11")
            file.write("\tstress_tensor_centrifugal_00")
            file.write("\tstress_tensor_centrifugal_01")
            file.write("\tstress_tensor_centrifugal_10")
            file.write("\tstress_tensor_centrifugal_11")
            file.write("\tstress_tensor_overall_00")
            file.write("\tstress_tensor_overall_01")
            file.write("\tstress_tensor_overall_10")
            file.write("\tstress_tensor_overall_11")

            p = self.plotter

            for ip in range(len(p.t)):
                file.write("\n" + str(ip))
                file.write("\t")
                file.write(str(p.t[ip]))
                file.write("\t")
                file.write(str(p.strain[ip]))
                file.write("\t")
                file.write(str(p.strain_rate[ip]))
                file.write("\t")
                file.write(str(p.force[ip]))
                file.write("\t")
                file.write(str(p.shear_stress[ip]))
                for i in range(2):
                    for j in range(2):
                        file.write("\t")
                        file.write(str(p.stress_tensor_LW[ip][i][j]))
                for i in range(2):
                    for j in range(2):
                        file.write("\t")
                        file.write(str(p.stress_tensor_linear_acceleration_otsuki[ip][i][j]))
                for i in range(2):
                    for j in range(2):
                        file.write("\t")
                        file.write(str(p.stress_tensor_with_external_forces[ip][i][j]))
                for i in range(2):
                    for j in range(2):
                        file.write("\t")
                        file.write(str(p.stress_tensor_linear_acceleration[ip][i][j]))
                for i in range(2):
                    for j in range(2):
                        file.write("\t")
                        file.write(str(p.stress_tensor_unbalanced_forces[ip][i][j]))
                for i in range(2):
                    for j in range(2):
                        file.write("\t")
                        file.write(str(p.stress_tensor_unbalanced_torque[ip][i][j]))
                for i in range(2):
                    for j in range(2):
                        file.write("\t")
                        file.write(str(p.stress_tensor_internal_torque[ip][i][j]))
                for i in range(2):
                    for j in range(2):
                        file.write("\t")
                        file.write(str(p.stress_tensor_spin_kinetic_energy[ip][i][j]))
                for i in range(2):
                    for j in range(2):
                        file.write("\t")
                        file.write(str(p.overall_stress_tensor[ip][i][j]))

        return [G_external[2],G_external[3]]













from tkinter import *
import random
import math
from .SphereLinkable import SphereLinkable
from .neighbor_relation_linkable import neighbor_relation_linkable
from particleShearBase import Force_register



class SphereLinkableAdjustableInterfaceStrength(SphereLinkable):
    def __init__(self, color, x, y, diameter, m=1,my_index=1, theCanvas=FALSE, doDrawing=FALSE,
                 force_register=Force_register(),size_x=500,size_y=500,permanent_ratio_central=1,permanent_ratio_tangential=1,
                 keep_viscosity_coefficients_constant=True):

        # The idea is here that we specifically modify the permanent links to have mechanical properties that can be different
        # from the purely frictional ones

        self.permanent_ratio_central=permanent_ratio_central
        self.permanent_ratio_tangential=permanent_ratio_tangential

        self.keep_viscosity_coefficients_constant=keep_viscosity_coefficients_constant





        super(SphereLinkableAdjustableInterfaceStrength,self).__init__(
            color, x, y, diameter, m,my_index, theCanvas, doDrawing,force_register,size_x,size_y)



    def tangential_force_elastic_permanent(self, theSphere, k_t):
        #print("tangential_force_elastic_permanent", k_t,self.permanent_ratio_tangential)
        return super(SphereLinkableAdjustableInterfaceStrength,self).\
            tangential_force_elastic_permanent(theSphere, k_t)*self.permanent_ratio_tangential

    def tangential_force_viscous_permanent(self, theSphere, nu):
        if(self.keep_viscosity_coefficients_constant):
            return super(SphereLinkableAdjustableInterfaceStrength, self).tangential_force_viscous_permanent(theSphere, nu)
        #print("tangential_force_viscous_permanent",self.permanent_ratio_tangential)
        return super(SphereLinkableAdjustableInterfaceStrength, self). \
                   tangential_force_viscous_permanent(theSphere, nu) * self.permanent_ratio_tangential


    def get_elastic_force_permanent(self,theSphere,k=1):
        #print("get_elastic_force_permanent",k,self.permanent_ratio_central)
        return super(SphereLinkableAdjustableInterfaceStrength, self). \
                   get_elastic_force_permanent(theSphere, k)*self.permanent_ratio_central


    def get_central_viscous_force(self, theSphere, nu):



        foundNeighborIndex = self.findNeighborIndex(theSphere)

        if(self.keep_viscosity_coefficients_constant):
            return super(SphereLinkableAdjustableInterfaceStrength, self).get_central_viscous_force(theSphere, nu)



        if (foundNeighborIndex >= 0):
            if (self.neighbors[foundNeighborIndex].interface_type != "permanent"):
                return super(SphereLinkableAdjustableInterfaceStrength, self).get_central_viscous_force(theSphere, nu)
            else:
                return super(SphereLinkableAdjustableInterfaceStrength, self).get_central_viscous_force(theSphere, nu)*\
                    self.permanent_ratio_central

        return super(SphereLinkableAdjustableInterfaceStrength, self).get_central_viscous_force(theSphere, nu)



    def contactLineColor(self,interface_type):
        col="green"
        if(interface_type == "stick" ):
            col="red"
        if (interface_type == "permanent" ):
            col = "darkorange"
        return col








# Basic friction-less sphere with drawing functions
class Graphical_output_configuration:
    """Provide configuration options for graphical output

    `particleShear.Graphical_output_configuration` objects are used to configure graphical representation of
     the spheres (class `particleShear.Circle` and derived).\n
     Sub-package particleShearBase
     """


    def __init__(self, draw_spheres_as="spheres",color_spheres_volume="grey",color_spheres_boundary="green",
                 draw_rotation_line=False, color_rotation_line="black",
                 draw_active_interface=True,color_interface_locked="red",color_interface_slip="green",
                 draw_permanent_connections=True,color_permanent_connections="blue",
                 draw_permanent_interfaces=False,
                 draw_volume_forces=False,color_volume_forces="darkorchid",
                 draw_boundary_forces=False,color_boundary_forces="darkorchid",
                 force_scale=1000,
                 draw_speeds=False,color_speeds="black"):
        """Initialization"""

        self.draw_spheres_as = draw_spheres_as
        """ Display of the spheres. Can be "spheres", "dots" or "none", 
        setting used by the spheres (Class `particleShear.Circle` and derived)"""
        self.color_spheres_volume = color_spheres_volume # Can be any valid python color, used by the spheres
        """  Drawing color of the spheres. 
        Can be any valid python color, used by the spheres (Class `particleShear.Circle` and derived)"""
        self.color_spheres_boundary = color_spheres_boundary # Can be any valid python color, used by the spheres
        """  Drawing color of the spheres on boundary. Not currently implemented. 
                Can be any valid python color, to be used by the spheres (Class `particleShear.Circle` and derived)"""
        self.draw_rotation_line = draw_rotation_line
        """ Boolean, indicates whehter the spheres to display rotation status (`particleShear.SphereFriction` and derived) """
        self.color_rotation_line = color_rotation_line # Can be any valid python color, used by the spheres
        """ Python color, indicates color for the rotation line (`particleShear.SphereFriction` and derived) """
        self.draw_active_interface = draw_active_interface
        """ Boolean, indicates whether currently active `particleShear.neighbor_relation`s should be shown 
        graphically. Used by the spheres 
        (`particleShear.CircleMassNeighbors` and derived)"""
        self.color_interface_locked = color_interface_locked
        """ Can be any valid python color, used by the spheres (`particleShear.CircleMassNeighbors` and derived). This 
        indicates the color for representation of the locked `particleShear.neighbor_relation`s"""
        self.color_interface_slip = color_interface_slip
        """ Can be any valid python color, used by the spheres (`particleShear.CircleMassNeighbors` and derived). This 
                indicates the color for representation of the locked `particleShear.neighbor_relation`s"""
        self.draw_permanent_connections = draw_permanent_connections
        """Boolean,  indicates to the spherees (`particleShear.SphereLinkable` and derived) whether permanent connections 
        should be drawn"""
        self.draw_permanent_interfaces=draw_permanent_interfaces
        """Boolean,  indicates to the spherees (`particleShear.SphereLinkable` and derived) whether the interfaces active 
        for the permanent connections should be drawn"""

        self.color_permanent_connections = color_permanent_connections
        """Valid Python color for the permanent connections, drawn from center to center
         """
        self.draw_volume_forces = draw_volume_forces
        """Boolean, used by `particleShear.StressTensorEvaluation` to show final forces on the
                non-constrained, totally interior spheres"""

        self.color_volume_forces = color_volume_forces
        """Any valid python color, used by `particleShear.StressTensorEvaluation`
         to show final forces on the non-constrained, interior spheres"""
        self.draw_boundary_forces = draw_boundary_forces
        """ Boolean, used by `particleShear.StressTensorEvaluation` to show final forces applied
            on the non-constrained spheres partially in contact with the constrained
            boundary spheres"""
        self.color_boundary_forces = color_boundary_forces
        """ Python color, used by `particleShear.StressTensorEvaluation` to show final forces applied
            on the non-constrained spheres partially in contact with the constrained
            boundary spheres"""
        self.force_scale = force_scale
        """ To graphically scale the forces to have reasonable length """
        self.draw_speeds = draw_speeds
        """Boolean, indicates whether spheres should draw their speed as vectors"""
        self.color_speeds = color_speeds
        """Python color, used by the spheres for the color of the speed vector"""




from tkinter import *
import random
import math
from .Sphere import Sphere
from particleShearBase import Force_register
from particleShearBase import CircleFrictionElasticity

class SphereFriction(CircleFrictionElasticity):
    """Objects of type `particleShear.CircleFrictionElasticity`, using standard rather than periodic Lees-Edwards
    boundary conditions"""
    def __init__(self, color, x, y, diameter, m=1,my_index=1, theCanvas=FALSE, doDrawing=FALSE,force_register=Force_register()):
        """Constructor as `particleShear.CircleFrictionElasticity.__init__` in `particleShear.CircleFrictionElasticity`
        but with and indexing element (my_index) and
        with the parameter use_lees_edwards set to False"""
        super(SphereFriction, self).__init__(color, x, y, diameter, m,  theCanvas, doDrawing, force_register,
                                             use_lees_edwards=False)


        self.myindex = my_index  # To give a unique index










from tkinter import *
import random
import math

from .Force_register import Force_register

from .neighbor_relation import neighbor_relation

from .Graphical_output_configuration import Graphical_output_configuration

from .CircleMassNeighbors import CircleMassNeighbors


def distance_transform_plateau(d, plateau_begin, plateau_end, relative_plateau_slope):
    """Helper function to transform the separation distance between particles to produce a force profile with
    a plateau, as typical for spongy materials

        - **parameters**\n
            `d` Center-center distance of the interacting particles\n
            `plateau_begin` Beginning of the plateau (smaller distance)\n
            `plateau_end` End of the plateau (larger distance)\n
            `relative_plateau_slope` Relative slope change on the plateau (1 = linear relation)\n
        """


    if d>=plateau_end:
        return d

    if d>=plateau_begin:
        return plateau_end + (d-plateau_end)*relative_plateau_slope

    return d/plateau_begin*(plateau_end + (plateau_begin-plateau_end)*relative_plateau_slope)


def elastic_force_law_plateau(d, d0, k, central_repulsion_coefficient):
    """Alternate elastic force law with a plateau at mid range of compression (for simulation of sponge-like materials)

    - **parameters**\n
        `d` Center-center distance of the interacting particles\n
        `d0` Equilibrium center-center distance\n
        `k` Spring constant\n
        `central_repulsion_coefficient` Possibility to have increased central repulsion at very closed distances.\n
    """

    if d > d0:
        return 0



    plateau_begin = PlateauConfiguration.relative_plateau_begin * d0
    plateau_end = PlateauConfiguration.relative_plateau_end * d0

    d = distance_transform_plateau(d,plateau_begin,plateau_end, PlateauConfiguration.relative_plateau_slope)


    linear_force = -k * (1 - central_repulsion_coefficient) * (d0 - d)
    nonlinear_addition = -k * central_repulsion_coefficient * d0 * (
                d0 / max(d, d0 / 1000) - 1)

    return linear_force + nonlinear_addition


def elastic_force_law(d, d0, k, central_repulsion_coefficient):
    """Default elastic force law with a plateau at mid range of compression (for simulation of sponge-like materials)

    Scalar value of the elastic force exerted on this sphere. A negative value signifies that the current object
    is pushed back by the other sphere, a positive value signifies attraction\n
    - **parameters**\n
        `d` Center-center distance of the interacting particles\n
        `d0` Equilibrium center-center distance\n
        `k` Spring constant\n
        `central_repulsion_coefficient` Possibility to have increased central repulsion at very closed distances.\n
    """

    if d > d0:
        return 0

    linear_force = -k * (1 - central_repulsion_coefficient) * (d0 - d)
    nonlinear_addition = -k * central_repulsion_coefficient * d0 * (
            d0 / max(d, d0 / 1000) - 1)

    return linear_force + nonlinear_addition


class PlateauConfiguration():
    """ Class to define alternative compressive force laws, not used by default"""
    relative_plateau_begin = 0.3
    relative_plateau_end = 0.8
    relative_plateau_slope = 0.1


class CircleBasicElasticity(CircleMassNeighbors):
    """Basic friction-less sphere with drawing functions.

    In the 2D approach chosen here, this a circle endowed with a central repulsive elastic force along with viscous
     dissipation if it comes into contact with another sphere.
    Inherits from `particleShear.PointLeesEdwards`, so will respect Lees-Edwards boundary conditions
    provided `particleShear.PointLeesEdwards.use_lees_edwards` is set to True (default: False)"""

    call_back_elastic_force_law = elastic_force_law
    """Possibility to pass a callback function to modify the force law"""

    def __init__(self, color, x, y, diameter, m=1,theCanvas=False, doDrawing=False,force_register=Force_register(),
                 use_lees_edwards=False):
        """Initialize self

        - **parameters**\n
        `x` The x-position of the center of the circle (in pixels, assumed micrometers for the simulation)\n
        `y` The y-position of the center of the circle (in pixels, assumed micrometers for the simulation)\n
        `diameter` The diameter of the circle  (in pixels, assumed micrometers for the simulation)\n
        `m` Mass of the sphere in mg (per m of depth). For the 2D representation, this is typically the mass in mg of a 1m high
        cylinder of diameter `diameter` in micrometers\n
        `theCanvas` Tkinter canvas; provide False if no Canvas is available or no drawing is desired\n
        `doDrawing` Explicit switch to turn of drawing\n
        `force_register`Possibility to provide a reference to a `particleShear.Force_register` variable to register the
        forces calculated during simulation\n
        `use_lees_edwards` Explicit Shoud Lees-Edwards boundary conditions be used?\n
        """

        super(CircleBasicElasticity,self).__init__(color,x,y,diameter,m=m, theCanvas=theCanvas,
                                                   doDrawing=doDrawing,force_register=force_register,
                                                   use_lees_edwards=use_lees_edwards)

        self.central_repulsion_coefficient=0
        """Central repulsion coefficient to help avoid spheres being able to fully penetrate and cross each other.

        The central_repulsion_coeffiecient is chosen from 0 (only linear law) to 1 
        (highest proportion of central repulsion). 
        When central_repulsion_coefficient>0, there is a 1/x contribution to induce very strong repulsion when 
        the spheres approach total compression and thus to avoid crossing of the spheres. For small compressions, 
        even if central_repulsion_coefficient>0, the equations are chosen to give a dF/dx law with a slope of -k 
        when the spheres just touch. See `particleShear.CircleBasicElasticity.get_elastic_force` further details"""






    #

    def get_elastic_force(self,theSphere,k=1):
        """Calculate magnitude of the central elastic force

        Scalar value of the elastic force exerted on this sphere. A negative value signifies that the current object
        is pushed back by the other sphere, a positive value signifies attraction (possible only with permanent
        interaction, not implemented in `particleShear.CircleBasicElasticity` but in some subclasses).\n
        Implementation of a central repulsive force: at large shear amplitudes and a strictly linear repulsion law,
        it may happen that spheres cross each other because the force is not enough to step movement despite the centers
        become nearly, or fully, identical. This is unphysical, and leads to entanglement between particles that we do
        no observe in this way. To avoid the effect, we need to add a component to the central elastic repulsive force that
        becomes very strong when the centers of the spheres get too close. At the same time, for small indentations, we
        aim at conserving a linear Hookes law (as in the paper underlying this implementation: see
        Eq. 2 from Otsuki, M. and H. Hayakawa,Discontinuous change of shear modulus for frictional
        jammed granular materials. Phys Rev E, 2017. 95(6-1): p. 062902.). Based on a 1/x law, we proceed as follows:\n
        The linear part is F=k*(x-xeq), providing the desired negative values for repulsion for x<xeq \n
        The nonlinear part is F=k*xeq*(1-xeq/x), which is also negative for xeq/x\n
        At x=xeq, both laws have dF/dx=k so that combining them in arbitrary mixing ratios does not change Hookes law
        for small indentations.\n
        This method is defined in class `particleShear.CircleBasicElasticity`"""


        d = self.d(theSphere)
        pos = theSphere.coordinates()
        r = pos[2]
        d0 = r + self.r

        return CircleBasicElasticity.call_back_elastic_force_law(d,d0,k,self.central_repulsion_coefficient)






    def elastic_force(self, theSphere, k):
        """Calculate and apply elastic force excerted on this sphere by `theSphere`

        This method will check for contact and so is more time
        consuming than `particleShear.CircleBasicElasticity.elastic_force_from_neighbors` \n
        This method is defined in class `particleShear.CircleBasicElasticity`"""
        d = self.d(theSphere)
        pos = theSphere.coordinates()
        r = pos[2]
        if d < r + self.r:
            force = self.get_elastic_force(theSphere,k)
            n_vector=self.n(theSphere)


            self.applyInternalForceAndRegister([force*n_vector[0], force*n_vector[1]],theSphere)






    def elastic_force_from_neighbors(self,k):
        """
        Calculate the central elastic force from known neighbors.

        For this function, the neighbor relations need to be already established (via
        `particleShear.CircleMassNeighbors.test_neighbor_relation`).\n
        This method is defined in class `particleShear.CircleBasicElasticity`"""
        for neighbor in self.neighbors:
            self.elastic_force(neighbor.theSphere,k)

    # Eq. 3 from Otsuki, M. and H. Hayakawa,
    # Discontinuous change of shear modulus for frictional
    # jammed granular materials. Phys Rev E, 2017. 95(6-1): p. 062902.

    def central_viscous_force(self, theSphere, nu):
        """Calculate and apply the central viscous force on this sphere by a given sphere (`theSphere`)

        Eq. 3 from Otsuki, M. and H. Hayakawa,
        Discontinuous change of shear modulus for frictional
        jammed granular materials. Phys Rev E, 2017. 95(6-1): p. 062902. This method will check for contact and so is more time
        consuming than `particleShear.CircleBasicElasticity.central_viscous_force_from_neighbors`\n
        This method is defined in class `particleShear.CircleBasicElasticity`.
        """

        force = self.get_central_viscous_force(theSphere,nu)

        n_vector = self.n(theSphere)
        self.xforce = self.xforce + force * n_vector[0]
        self.yforce = self.yforce + force * n_vector[1]
        self.force_register.record_individual_internal_force(self,theSphere,[force * n_vector[0], force * n_vector[1]])

        return [force * n_vector[0], force * n_vector[1]]

    def get_central_viscous_force(self, theSphere, nu):
        """Calculate and apply the central viscous force on this sphere by a given sphere (`theSphere`)

        Eq. 3 from Otsuki, M. and H. Hayakawa,
        Discontinuous change of shear modulus for frictional
        jammed granular materials. Phys Rev E, 2017. 95(6-1): p. 062902. The magnitude of the viscous force depends
        on the relative speed. The convention is similar to the central elastic force (
        `particleShear.CircleBasicElasticity.get_elastic_force`). If the current object is
        repulsed by an approaching sphere (`theSphere`), a negative force is returned.\n
        This method is defined in class `particleShear.CircleBasicElasticity`.
        """
        v_rel = self.relative_speed(theSphere)

        vx_rel = v_rel[0]
        vy_rel = v_rel[1]

        norm_vector = self.n(theSphere)
        v_norm = vx_rel * norm_vector[0] + vy_rel * norm_vector[1]

        force = nu * v_norm

        return force


    def central_viscous_force_from_neighbors(self, nu):
        """Calculate the central viscous force from known neighbors.

        For this function, the neighbor relations need to be already established (via
        `particleShear.CircleMassNeighbors.test_neighbor_relation`).\n
        This method is defined in class `particleShear.CircleBasicElasticity`"""
        for neighbor in self.neighbors:
            self.central_viscous_force(neighbor.theSphere,nu)







from particleShearObjects import particle_shear_model_parameters
from .OscillatoryShearExperiment import OscillatoryShearExperiment
from .OscillatorySimulation import OscillatorySimulation
import math
from tkinter import *
from pathlib import Path



class OscillatorySimulationFragmentation(OscillatorySimulation):

    def __init__(self,root_folder,do_permanent_links=True,cut_lines=5,N=150,packing_fraction=2,amplitude=0.5,
                 Young_modulus_spheres=1000,mu=0.1,
                 relative_y_scale_force=5e-4,relative_frequency=0.05,doDrawing=True,saveData=False):
        self.packing_fraction = packing_fraction
        self.N = N

        super(OscillatorySimulationFragmentation,self).__init__(
            root_folder,amplitude=amplitude,
            relative_y_scale_force=relative_y_scale_force,relative_frequency=relative_frequency,doDrawing=doDrawing,
            saveData=saveData)
        self.do_permanent_links=do_permanent_links
        self.cut_lines=cut_lines



        self.Young_modulus_spheres=Young_modulus_spheres
        self.mu=mu
        self.theEnsemble=False

        self.model=False

        self.packing_fraction=packing_fraction

        self.function_call="OscillatorySimulationFragmentation(root_folder="+str(root_folder)+",do_permanent_links="+\
            str(do_permanent_links)+",cut_lines="+str(cut_lines)+",N="+str(N)+\
            "packing_fraction="+str(packing_fraction)+",amplitude="+str(amplitude)+\
            ",Young_modulus_spheres="+str(Young_modulus_spheres)+",mu="+str(mu)+\
            ",relative_y_scale_force="+str(relative_y_scale_force)+",relative_frequency="+str(relative_frequency)+\
            ",doDrawing=" + str(doDrawing) + ")"

    def adjustment_factor_for_bounded_preequilibration_steps(self):
        if self.mu > 0 and self.mu<=1:
            return int(2 - math.log10(self.mu))
        return 1

    # Here the ensemble needs to be initiated, but this needs to be done in the subclasses
    def initEnsemble(self):

        if self.doDrawing and not self.theTkSimulation:
            self.theTkSimulation=Tk()



    def file_path(self):
        if not self.root_folder:
            return False
        fname = "dermal_filler"

        if self.do_permanent_links:
            if not (self.cut_lines > 0):
                fname = fname + "_bulk"
        else:
            fname = fname + "_uncrosslinked"

        fname = fname + str(self.mu) + "_f_" + str(round(self.f*10000)/10000) + "_strain_" +\
                str(round(self.amplitude*100000)/100000) + "_nr_"

        fnum = 1

        while Path(self.root_folder + "/" + fname + str(fnum) + ".txt").is_file():
            fnum = fnum + 1

        return self.root_folder + "/" + fname + str(fnum) + ".txt"

    def write_information_on_simulation_to_file(self,theFile):
        if not self.saveData:
            return


        if self.do_permanent_links:
            if self.cut_lines > 0:
                theFile.write("Dermal filler Ensemble: \n")
                theFile.write("\tCutting lines used to generate the particles" + str(self.cut_lines) + "\n")
            else:
                theFile.write("Dermal filler control ensemble: no cut lines \n")
        else:
            theFile.write("Dermal filler control ensemble: No permanent links\n")

        theFile.write("\tN=" + str(self.N) + " spheres, nominal packing density = " + str(self.packing_fraction) + "\n")

        theFile.write("\tYoung modulus of the individual spheres = " + str(self.Young_modulus_spheres) + " Pa \n")

        theFile.write("\tFriction coefficient = " + str(self.mu) + "\n\n")

        theFile.write("\n\nFunction call : \n" + self.function_call+"\n\n")






from tkinter import *
import random
import math
from .Force_register import Force_register
from .Graphical_output_configuration import Graphical_output_configuration
from .CanvasPointsMass import CanvasPointsMass


class CanvasPointsNeighbors(CanvasPointsMass):
    """Canvas for placing `particleShear.CircleMassNeighbors` objects (referred to as spheres).
           This class maintains a list of objects of type
           `particleShear.CircleMassNeighbors` or derived and adds basic functions to handle this collection"""

    def __init__(self, size_x, size_y, theCanvas=False, doDrawing=False, m=1):

        super(CanvasPointsNeighbors,self).__init__(size_x=size_x,size_y=size_y,theCanvas=theCanvas,doDrawing=doDrawing,m=m)




    def particle_info(self):
        return "CanvasPointsNeighbors: particles with mass and detection of geometric neighbors"

    def test_neighbor_relation(self):
        """Instruct every sphere to test establish its geometric (and possibly permanent) neighbors.

        This function calls `particleShear.CircleMassNeighbors.test_neighbor_relation` on each member of the
        `particleShear.CanvasPoints.sphereList`.\n
         This method is defined in class `particleShear.CanvasPointsNeighbors`"""
        for sphereIndex in range(len(self.sphereList)):
            for sphereIndex2 in range(len(self.sphereList)):

                if sphereIndex != sphereIndex2:
                    self.sphereList[sphereIndex].test_neighbor_relation(
                        self.sphereList[sphereIndex2]
                    )











from particleShearBase import neighbor_relation


# Object holding the information to the neighboring spheres
class neighbor_relation_linkable(neighbor_relation):
    def __init__(self,friction_position,interface_type,theSphere,graphical_line_index=-1,linking_line_index=-1):
        super(neighbor_relation_linkable,self).__init__(friction_position,interface_type,theSphere,
                                                        graphical_line_index=graphical_line_index)
        self.linking_line_index = linking_line_index  # -1 is for no linking line yet
        self.equilibrium_distance=0





from tkinter import *
import random
import math

from .Force_register import Force_register

from .Graphical_output_configuration import Graphical_output_configuration

from .Circle import Circle


class CircleMass(Circle):
    """Provide a circle with an attribued mass

    Describes a sphere (circle in 2D) that possess a mass and stores the forces acting on the sphere. \n
    In a simulation, the concept is that before each
     simulation step, the force known to each sphere itself (the ìnstance variables `particleShear.CircleMass.xforce` and
    `particleShear.CircleMass.yforce`) are reset to 0; as the spheres are meant to be associated with a simulation
     area object derived from `particleShear.CanvasPointsMass`, this typically happens through a call to the method
     `particleShear.CanvasPointsMass.reset_force` of `particleShear.CanvasPointsMass` or relevant subclass.
     In addition, for the purpose of stress tensor evaluation, details about
     the forces are registered during the various calculations in a simulation step in the
     `particleShear.CircleMass.force_register`, which is provided through the canvas class (derived from
     `particleShear.CanvasPointsMass`). At the beginning of each simulation step, the force register common to all the
     spheres on a simulaton canvas also needs to be reset, this is done via the
     `particleShear.CanvasPointsMass.reset_force_register` of `particleShear.CanvasPointsMass` or relevant subclass.\n
     When running the simulation, forces are calculated in various ways, and attributed to the sphere particles by direct
     change of the instance variables `particleShear.CircleMass.xforce` and
    `particleShear.CircleMass.yforce`. Depending on their nature, they are also registered in various ways in the common
     `particleShear.Force_register` object,
     which is the instance variable `particleShear.CanvasPointsMass.force_register` of the relevant subclass of
     `particleShear.CanvasPointsMass`; for practical purposes, this very same `particleShear.Force_register` is also known
     to each sphere as its own instance variable `particleShear.CircleMass.force_register`.\n
    Once all the forces known after a complete simulation step, the net force can be
    converted to acceleration (by explicit call to `particleShear.CircleMass.do_linear_acceleration`). This then
    calculates the change in speed due to the acceleration, and resets the force for the particle at hand to zero.
    At this point, the `particleShear.CircleMass.force_register`, common to all the spheres,
    still contains the detailed information about the forces and can
    be exploited for stress tensor evaluation by a `particleShear.StressTensorEvaluation` object. The
    `particleShear.CircleMass.force_register` needs to be reset separately prior to the next simulation step, see above.\n
    Inherits from `particleShear.PointLeesEdwards`, so will respect Lees-Edwards boundary conditions
    provided `particleShear.PointLeesEdwards.use_lees_edwards` is set to True (default: False)"""

    def __init__(self, color, x, y, diameter, m=1, theCanvas=False, doDrawing=False,force_register=Force_register(),
                 use_lees_edwards=False):
        """Initialize self

        - **parameters**\n
        `x` The x-position of the center of the circle (in pixels, assumed micrometers for the simulation)\n
        `y` The y-position of the center of the circle (in pixels, assumed micrometers for the simulation)\n
        `diameter` The diameter of the circle  (in pixels, assumed micrometers for the simulation)\n
        `m` is the mass per unit of depth in mg/m of depth\n
        `theCanvas` Tkinter canvas; provide False if no Canvas is available or no drawing is desired\n
        `doDrawing` Explicit switch to turn of drawing\n
        `force_register`Possibility to provide a reference to a common `particleShear.Force_register` variable to register the
        forces calculated during simulation\n
        `use_lees_edwards` Explicit Shoud Lees-Edwards boundary conditions be used?\n
        """


        super(CircleMass,self).__init__(color,x,y,diameter, theCanvas=theCanvas, doDrawing=doDrawing,
                                        use_lees_edwards=use_lees_edwards)

        self.m = m
        """Mass of the particle, in mg/m of depth (2D simulaton)"""


        self.force_register = force_register
        """Reference to the common `particleShear.CanvasPointsMass.force_register` to store the forces acting on the spheres"""

        self.xforce = 0
        """x-component of the net force on this sphere, in N per m of depth in the 2D simulation"""
        self.yforce = 0
        """y-component of the net force on this sphere, in N per m of depth in the 2D simulation"""


    def record_total_particle_force(self):
        """ Record the current total force on this particle

        This functions transmits the current net force having served to do the
        calculation to the `particleShear.CircleMass.force_register`

        This method is defined in class `particleShear.CircleMass`"""

        self.force_register.record_total_particle_force(self, [self.xforce, self.yforce])

    def do_linear_acceleration(self, dt):
        """ Apply the linear acceleration from the net force

        This method calculates the linear acceleration from F=m*a and applies it by dv=a*dt. It also transmits
        the current total force to the force register (via `particleShear.CircleMass.record_total_particle_force`) before
        resetting the net force to 0\n
        This method is defined in class `particleShear.CircleMass`"""
        self.record_total_particle_force()

        self.xspeed = self.xspeed + self.xforce * dt / self.m
        self.yspeed = self.yspeed + self.yforce * dt / self.m
        self.xforce = 0
        self.yforce = 0

    def applyInternalForceAndRegister(self,force_vector,source,register_homolog=False):
        """ Apply an internal force and update the `particleShear.CircleMass.force_register`

        This method adds a given force_vector to the net force acting on this particle.
        It also registers the force in the `particleShear.CircleMass.force_register` as an internal force
        (between two mobile spheres). If the `register_homolog` parameter is provided as True,
        then the opposing equal force is also registered  in the `particleShear.CircleMass.force_register`. Even in that
        case, the opposing equal reaction force has to be explicity applied on the `source` sphere elsewhere, this
        function never applies a force directly to the `source` sphere.\n
        This method is defined in class `particleShear.CircleMass`"""

        self.xforce = self.xforce + force_vector[0]
        self.yforce = self.yforce + force_vector[1]

        self.force_register.record_individual_internal_force(self, source,force_vector)
        if register_homolog:
            opposing_force = [-force_vector[0], -force_vector[1]]
            source.force_register.record_individual_internal_force(source, self, opposing_force)









from tkinter import *
import random
import math

from .PointLeesEdwards import PointLeesEdwards
from .Graphical_output_configuration import Graphical_output_configuration


#
class Circle(PointLeesEdwards):
    """Provide a circle that can draw itself according to the graphical drawing convention

        Inherits from `particleShear.PointLeesEdwards`, so will respect Lees-Edwards boundary conditions
        provided `particleShear.PointLeesEdwards.use_lees_edwards` is set to True (default: False)"""

    def __init__(self, color, x, y, diameter, theCanvas=False, doDrawing=False,use_lees_edwards=False):
        """Initialize self

        - **parameters**\n
        `x` The x-position of the center of the circle (in pixels, assumed micrometers for the simulation)\n
        `y` The y-position of the center of the circle (in pixels, assumed micrometers for the simulation)\n
        `diameter` The diameter of the circle  (in pixels, assumed micrometers for the simulation)\n
        `theCanvas` Tkinter canvas; provide False if no Canvas is available or no drawing is desired\n
        `doDrawing` Explicit switch to turn of drawing\n
        `use_lees_edwards` Explicit Shoud Lees-Edwards boundary conditions be used?\n
        """
        super(Circle,self).__init__(x,y,use_lees_edwards=use_lees_edwards)

        self.theCanvas = theCanvas
        """tkinter canvas"""

        self.r = diameter / 2
        """Radius of the circle, in pixels (=micrometers for the simulations)"""
        self.doDrawing = doDrawing
        """Flag to indicate whether graphical output should be produced or not"""

        self.shape=False
        """Shape on the Tkinter canvas when actively drawing"""
        self.color=color
        """Color of the spheres"""
        self.graphical_output_configuration = Graphical_output_configuration()
        """`particleShear.Graphical_output_configuration` to set options how the spheres and their neighboring relations
                are displayed"""
        self.graphical_output_configuration.color_spheres_volume= color
        self.graphical_output_configuration.color_spheres_boundary = color


        self.set_graphical_output_configuration(self.graphical_output_configuration)

        self.initiate_drawing()




    def set_graphical_output_configuration(self, graphical_output_configuration):
        """Set new `particleShear.Graphical_output_configuration`

        This method is defined in class `particleShear.Circle`"""
        self.deleteDrawing()
        self.graphical_output_configuration=graphical_output_configuration
        self.color=self.graphical_output_configuration.color_spheres_volume
        self.initiate_drawing()

    def coordinates(self):
        """Return the current coordinates: vector of three elements, which are (x,y,radius)

        This method is defined in class `particleShear.Circle`"""
        return [self.x, self.y, self.r]


    def initiate_drawing(self):
        """Instiante the shapes representing this object on  `particleShear.Circle.theCanvas`.

        This method only has an effect if `particleShear.Circle.doDrawing` is True. If a shape is created, the reference to
        the circle on the canvas is stored in `particleShear.Circle.shape`"""
        self.deleteDrawing()

        self.shape = False

        if self.doDrawing:
            if self.graphical_output_configuration.draw_spheres_as=="spheres":
                self.shape = self.theCanvas.create_oval(self.x - self.r, self.y - self.r, self.x + self.r, self.y + self.r,
                                               fill=self.color)
            if self.graphical_output_configuration.draw_spheres_as=="dots":
                self.shape = self.theCanvas.create_oval(self.x - 2, self.y - 2, self.x + 2,
                                                        self.y + 2,
                                                        fill=self.color)


    def move(self, dt=1):
        """Move for time interval dt, including the graphical shape if present

        This method is defined in class `particleShear.Circle`"""
        super(Circle,self).move(dt)
        if self.doDrawing:
            self.moveDrawing(dt)




    def moveDrawing(self, dt):
        """Move for the drawing on the Tkinter canvas

        This method is defined in class `particleShear.Circle`"""
        self.theCanvas.move(self.shape, self.xspeed * dt, self.yspeed * dt)

    def deleteDrawing(self):
        """Delete drawing on the Tkinter canvas

        This method is defined in class `particleShear.Circle`"""
        if self.theCanvas:
            self.theCanvas.delete(self.shape)

    def tangential_speed(self, theSphere):
        """Tangential component of the relative speed of a given sphere relative to this object.

        The other sphere object needs to be of type `particleShear.Point` or derived. Returns a scalar,
        describing the relative speed component parallel to a tangential vector with counter-clockwise orientation.
        The speed is returned in pixels/s (= micrometers/s for the simulation). This method uses
        `particleShear.Circle.tangential_vector` internally.\n
         This method is defined in class `particleShear.Circle`"""
        v_rel = self.relative_speed(theSphere)
        vx_rel = v_rel[0]
        vy_rel = v_rel[1]

        t_vector = self.tangential_vector(theSphere)
        vt = vx_rel * t_vector[0] + vy_rel * t_vector[1]
        return vt


    def tangential_vector(self, theSphere):
        """Return tangential interface vector of given sphere relative to this object.

        Calculates the tangential vector describing the direction and orientation of interface located perpendicalurlay
        to the line relying the centers of the spheres. Counter-clock wise orientation (but: attention to the
        orientation of the computer graphics device, which often invert the y-axis!)\n
         This method is defined in class `particleShear.Circle`"""
        n_vector = self.n(theSphere)
        return [-n_vector[1], n_vector[0]]








from tkinter import *
import random
import math
from particleShearObjects import SphereFrictionLeesEdwards
from .neighbor_relation_linkable import neighbor_relation_linkable
from particleShearBase import Force_register
from particleShearBase import CircleBasicElasticity
from math import exp


def elastic_force_law_tensile(d, d0, k):
    return -k * (d0 - d)

def elastic_force_law_tensile_exponential(d, d0, k):
    return k*(1-exp(-(d-d0)/d0/TensileConfiguration.d_exponential))*d0*TensileConfiguration.d_exponential






class TensileConfiguration():
    """ Class to define alternative tensile force laws, not used by default"""

    d_exponential=0.2


class SphereLinkable(SphereFrictionLeesEdwards):
    call_back_elastic_force_law_tensile = elastic_force_law_tensile


    def __init__(self, color, x, y, diameter, m=1,my_index=1, theCanvas=FALSE, doDrawing=FALSE,
                 force_register=Force_register(),size_x=500,size_y=500):
        super(SphereLinkable,self).__init__(color, x, y, diameter, m,my_index, theCanvas, doDrawing,force_register,size_x,size_y)


    def cut_permanent_link(self,theSphere,do_backlink=True):
        foundNeighborIndex = self.findNeighborIndex(theSphere)

        if foundNeighborIndex >= 0:
            theNeighbor = self.neighbors[foundNeighborIndex]
            if theNeighbor.interface_type=="permanent":

                theNeighbor.equilibrium_distance = 0
                theNeighbor.friction_position = 0
                theNeighbor.interface_type = "slip"

                # Link back from the neighbor
                if do_backlink:
                    theNeighbor.theSphere.cut_permanent_link(self, do_backlink=False)
                if self.doDrawing:
                    self.initiateNeighborDrawing()

    def establish_permanent_link(self,theSphere,do_backlink=True):
        foundNeighborIndex = self.findNeighborIndex(theSphere)

        if foundNeighborIndex>=0:
            theNeighbor = self.neighbors[foundNeighborIndex]
            theNeighbor.equilibrium_distance=self.r+theNeighbor.theSphere.r
            theNeighbor.friction_position=0
            theNeighbor.interface_type="permanent"

            # Link back from the neighbor
            if do_backlink:
                theNeighbor.theSphere.establish_permanent_link(self,do_backlink=False)
            if self.doDrawing:

                self.initiateNeighborDrawing()

        else:
            info = neighbor_relation_linkable(0, "permanent", theSphere)
            info.equilibrium_distance=self.r+theSphere.r
            info.friction_position = 0



            if self.doDrawing:
                self.deleteNeighborDrawing()
            self.neighbors.append(info)

            if do_backlink:
                theSphere.establish_permanent_link(self,do_backlink=False)


            if self.doDrawing:
                self.initiateNeighborDrawing()

    def permanentlyConnectedSpheres(self,physical_neighbors_only=True):
        connected=[]
        for theNeighbor in self.neighbors:
            if theNeighbor.interface_type=="permanent" and not connected.count(theNeighbor.theSphere):
                if physical_neighbors_only: # Additional check for physical proximity
                    if abs(theNeighbor.theSphere.x-self.x)<self.size_x*0.5 and abs(theNeighbor.theSphere.y-self.y)<self.size_y*0.5:
                        connected.append(theNeighbor.theSphere)
        return connected








    def test_neighbor_relation(self, theSphere):
        d = self.d(theSphere)
        pos = theSphere.coordinates()
        r = pos[2]
        is_neighbor=(d < r + self.r)
        # Regardless of the distance, first check whether we find the sphere in the neighbors list
        found = FALSE
        foundNeighborIndex=-1
        for theNeighborIndex in range(len(self.neighbors)):
            theNeighbor = self.neighbors[theNeighborIndex]
            if (theSphere.myindex == theNeighbor.myindex):
                foundNeighborIndex = theNeighborIndex
                found = TRUE

        if not is_neighbor: #If the sphere is not a neighbor (anymore)
            if found: #It was previously a neighbor, but no more, so delete it from the neighbors list if possible
                if not self.neighbors[foundNeighborIndex].interface_type=="permanent":
                    if self.doDrawing:
                        self.deleteNeighborDrawing()
                    del self.neighbors[foundNeighborIndex]
                    if self.doDrawing:
                        self.initiateNeighborDrawing()

        if is_neighbor:
            if not found: # Not found previously, so add a new entry
                info = neighbor_relation_linkable(0,"stick",theSphere)
                if self.doDrawing:
                    self.deleteNeighborDrawing()
                self.neighbors.append(info)
                if self.doDrawing:
                    self.initiateNeighborDrawing()



    def deleteNeighborDrawing(self):
        super(SphereLinkable,self).deleteNeighborDrawing()
        for theNeighbor in self.neighbors:
            if theNeighbor.linking_line_index!=-1:
                if self.theCanvas:
                    self.theCanvas.delete(theNeighbor.linking_line_index)
                theNeighbor.linking_line_index=-1

    def initiateNeighborDrawing(self):
        super(SphereLinkable, self).initiateNeighborDrawing()
        for theNeighbor in self.neighbors:
            if theNeighbor.linking_line_index==-1:
                if theNeighbor.interface_type == "permanent":
                    # To correctly draw under periodic boundary conditions we need to use the joining vector and distance
                    # rather than raw coordinates
                    dist=self.d(theNeighbor.theSphere)
                    n_vector = self.n(theNeighbor.theSphere)
                    if self.theCanvas:
                        theNeighbor.linking_line_index = self.theCanvas.create_line(
                            self.x,
                            self.y,
                            self.x + n_vector[0] * dist,
                            self.y + n_vector[1] * dist,
                            width=3, fill="blue"
                        )

    # Here, permanent and transient interfaces behave the same. This is not necessarily true for subclasses
    def tangential_force_viscous_permanent(self,theSphere,nu):
        return self.tangential_force_viscous(theSphere,nu)



    def tangential_force(self, theSphere, nu,mu,k,k_t):
        d = self.d(theSphere)
        pos = theSphere.coordinates()
        r = pos[2]

        foundNeighborIndex=self.findNeighborIndex(theSphere)

        found = foundNeighborIndex >= 0

        isPermanent = False






        if found:

            # Check whether symmetric backlink exists

            backlink=self.neighbors[foundNeighborIndex].theSphere.findNeighborIndex(self)

            if(not backlink>=0):
                print("SphereLinkable: Asymmetric neighbor problem")



            if self.neighbors[foundNeighborIndex].interface_type == "permanent":
                isPermanent = True



        if d < r + self.r or isPermanent:


            if not found:
                print("tangential_force: Problem with neighbors, found touching sphere not in neighbors list")
                return




            if isPermanent:
                viscous_force=self.tangential_force_viscous_permanent(theSphere,nu)
                if abs(viscous_force - theSphere.tangential_force_viscous_permanent(self, nu)) > 1e-15:
                    print("viscous force mis-match", viscous_force - theSphere.tangential_force_viscous_permanent(self, nu))

            else:
                viscous_force=self.tangential_force_viscous(theSphere,nu)
                if abs(viscous_force - theSphere.tangential_force_viscous(self, nu)) > 1e-15:
                    print("viscous force mis-match", viscous_force - theSphere.tangential_force_viscous(self, nu))












            if isPermanent:
                elastic_force=self.tangential_force_elastic_permanent(theSphere,k_t)
            else:
                elastic_force=self.tangential_force_elastic(theSphere,k_t)






            total_adherence_force = viscous_force+elastic_force
            friction_force=self.get_elastic_force(theSphere, k)*mu






            sig=1
            if(total_adherence_force<0):
                sig=-1

            if(abs(total_adherence_force)<=abs(friction_force)):
                if not isPermanent:
                    self.neighbors[foundNeighborIndex].interface_type="stick"
                    theSphere.neighbors[backlink].interface_type = "stick"
            else:
                if not isPermanent:
                    self.neighbors[foundNeighborIndex].interface_type = "slip"
                    self.neighbors[foundNeighborIndex].friction_position=0
                    theSphere.neighbors[backlink].interface_type = "slip"
                    theSphere.neighbors[backlink].friction_position = 0
            if isPermanent:
                force=abs(total_adherence_force)
            else:
                force=min(abs(total_adherence_force),abs(friction_force))



            force=force*sig

            # The idea here is that the same friction point will be visited again when the neighboring sphere is examined,
            # so we assign only half the force here
            # The distribution is a bit complicated because of the conservation of angular momentum, so we use a dedicated function
            self.distribute_tangential_couple(theSphere, force/2)


    def tangential_force_elastic_permanent(self, theSphere, k_t):
        found = False
        foundNeighborIndex = -1
        for theNeighborIndex in range(len(self.neighbors)):
            theNeighbor = self.neighbors[theNeighborIndex]
            if (theSphere == theNeighbor.theSphere and theNeighbor.interface_type=="permanent"):
                foundNeighborIndex = theNeighborIndex
                found = TRUE
        if not found:
            print("Problem with neighbors, found permanent touching sphere not in neighbors list")
        else:

            return k_t * self.neighbors[foundNeighborIndex].friction_position

    def elastic_force(self, theSphere, k):



        foundNeighborIndex = self.findNeighborIndex(theSphere)

        if(foundNeighborIndex>=0):
            if(self.neighbors[foundNeighborIndex].interface_type != "permanent"):
                super(SphereLinkable,self).elastic_force(theSphere,k)
                return

        # So, we have a permanent neighbor here with a known equilibrium position

        d = self.d(theSphere)
        pos = theSphere.coordinates()
        r = pos[2]

        force = self.get_elastic_force_permanent(theSphere,k)

        n_vector = self.n(theSphere)

        self.xforce = self.xforce + force * n_vector[0]
        self.yforce = self.yforce + force * n_vector[1]


        self.force_register.record_individual_internal_force(self,theSphere,[force * n_vector[0], force * n_vector[1]])



    def get_elastic_force_permanent(self,theSphere,k=1):




        d = self.d(theSphere)

        foundNeighborIndex = self.findNeighborIndex(theSphere)

        if(foundNeighborIndex<0):
            print("Could not find supposedly permanent neighbor")
            return 0

        equilibrium_distance = self.neighbors[foundNeighborIndex].equilibrium_distance

        if d>=equilibrium_distance:

            return SphereLinkable.call_back_elastic_force_law_tensile(d, equilibrium_distance, k)




        return CircleBasicElasticity.call_back_elastic_force_law(d, equilibrium_distance, k, self.central_repulsion_coefficient)












    def contactLineColor(self,interface_type):
        col="green"
        if(interface_type == "stick" or interface_type == "permanent"):
            col="red"
        return col







from tkinter import *
import random
import math
from .Force_register import Force_register
from .Graphical_output_configuration import Graphical_output_configuration
from .CanvasPointsShear import CanvasPointsShear



class CanvasPointsBasicElasticity(CanvasPointsShear):
    """Canvas for placing `particleShear.CircleBasicElasticity` objects (referred to as spheres).
       This class maintains a list of objects of type
       `particleShear.CircleBasicElasticity` or derived and adds basic functions to handle this collection"""

    def __init__(self, size_x, size_y, theCanvas=False, doDrawing=False, k=1, nu=0.01,m=1):
        """Initialize self

                - **parameters**\n
                    `size_x` Width of area to be used in pixels = micrometers for the simulation\n
                    `size_y` Height of area to be used in pixels = micrometers for the simulation\n
                    `theCanvas` possibility to transmit a `tkinter` `Canvas` object for graphical output\n
                    `doDrawing` Flag to indicate whether graphical output should be produced or not\n
                    `k` Spring constant in (mg/s^2)/m of depth;
                    so F=-k*x*L with L the depth of the stack (the simulation is 2D) and x the compression\n
                    `nu` Viscosity force constant in (mg/s)/m of depth. This is to have F=nu*v*L, L again the depth\n
                    `m` is the mass per unit of depth in mg/m of depth"""

        self.k = k
        """Central spring constant in (mg/s^2)/m of depth"""
        self.nu = nu
        """Central viscosity constant in (mg/s)/m of depth"""

        self.central_repulsion_coefficient=0
        """Central repulsion coefficient to help avoid spheres being able to fully penetrate and cross each other.
         
        The central_repulsion_coeffiecient is chosen from 0 (only linear law) to 1 
        (highest proportion of central repulsion). 
        When central_repulsion_coefficient>0, there is a 1/x contribution to induce very strong repulsion when 
         the spheres approach total compression and thus to avoid crossing of the spheres. For small compressions, 
         even if central_repulsion_coefficient>0, the equations are chosen to give a dF/dx law with a slope of -k 
         when the spheres just touch. See `particleShear.CircleBasicElasticity.get_elastic_force` for details"""

        super(CanvasPointsBasicElasticity,self).__init__(size_x=size_x,size_y=size_y,theCanvas=theCanvas,doDrawing=doDrawing,m=m)




    def particle_info(self):
        return "CanvasPointsBasicElasticity: Basic elastic and neighbor behaviour"



    def elastic_force(self):
        """Let all spheres calculate the central elastic force acting on them; this will check for contact and so is more time
        consuming than `particleShear.CanvasPointsBasicElasticity.elastic_force_from_neighbors`

        This method is defined in class `particleShear.CanvasPointsBasicElasticity`"""
        for sphereIndex in range(len(self.sphereList)):
            for sphereIndex2 in range(len(self.sphereList)):
                if sphereIndex != sphereIndex2:
                    self.sphereList[sphereIndex].elastic_force(self.sphereList[sphereIndex2], self.k)


    def elastic_force_from_neighbors(self):
        """Let all spheres calculate the central elastic force acting on from their known neighbors.

           This is quicker than `particleShear.CanvasPointsBasicElasticity.elastic_force`
           but necessitates the neighbor relations to be already established via
           `particleShear.CanvasPointsNeighbors.test_neighbor_relation`\n
           This method is defined in class `particleShear.CanvasPointsBasicElasticity`"""


        for sphereIndex in range(len(self.sphereList)):
            self.sphereList[sphereIndex].elastic_force_from_neighbors(self.k)

    def central_viscous_force(self):
        """Let all spheres calculate the central viscous force acting on them.

        This will check for contact and so is more time
        consuming than `particleShear.CanvasPointsBasicElasticity.central_viscous_force_from_neighbors`.\n
        This method is defined in class `particleShear.CanvasPointsBasicElasticity`"""
        for sphereIndex in range(len(self.sphereList)):
            for sphereIndex2 in range(len(self.sphereList)):
                if sphereIndex != sphereIndex2:
                    self.sphereList[sphereIndex].central_viscous_force(self.sphereList[sphereIndex2],
                                                                       self.nu)

    def central_viscous_force_from_neighbors(self):
        """Let all spheres calculate the central viscous force acting on from their known neighbors.

            This is quicker than `particleShear.CanvasPointsBasicElasticity.central_viscous_force`
            but necessitates the neighbor relations to be already established via
            `particleShear.CanvasPointsNeighbors.test_neighbor_relation`\n
            This method is defined in class `particleShear.CanvasPointsBasicElasticity`"""
        for sphereIndex in range(len(self.sphereList)):
            self.sphereList[sphereIndex].central_viscous_force_from_neighbors(self.nu)


    def mechanical_simulation_step(self, cool_factor=0.97, dt=1):
        """ Perform full mechanical simulation step.

            This corresponds to the events associated with time advancing by dt.
            For this class, this includes resetting all forces to zero, calculation and summation of
            central and tangential forces on all particles, and calculation of the resulting
            acceleration, rotational acceleration, linear and rotational movement, including Lees-Edwards boundary
            conditions and adaption to change of shear rate
            The step can also include cooling (translational and rotational) by cool_factor.
            Forces are registered in the `particleShear.CanvasPointsMass.force_register`\n
            This method is defined in class `particleShear.CanvasPointsBasicElasticity`"""

        self.mechanical_simulation_step_calculate_forces()
        self.record_total_particle_forces()
        self.mechanical_simulation_step_calculate_acceleration(cool_factor=cool_factor,dt=dt)
        self.mechanical_simulation_step_calculate_movement(dt=dt)



    def mechanical_simulation_step_calculate_forces(self):
        """ Calculate the forces acting on the spheres

        This method is defined in class `particleShear.CanvasPointsBasicElasticity`"""
        self.reset_force_register()
        self.test_neighbor_relation()
        self.elastic_force_from_neighbors()
        self.central_viscous_force_from_neighbors()



    def mechanical_simulation_step_calculate_acceleration(self,cool_factor=0.97,dt=1):
        """Calculate the accelerations resulting from the forces

        This method is defined in class `particleShear.CanvasPointsBasicElasticity`"""
        self.do_linear_acceleration(dt)
        self.cool(cool_factor)


    def mechanical_simulation_step_calculate_movement(self,dt=1):
        """ Calculate the displacement resulting from the sphere movement

        This method is defined in class `particleShear.CanvasPointsBasicElasticity`"""
        self.move(dt)
        self.boundary_conditions()
        self.t = self.t + dt
        if self.applyingShear:
            self.shear = self.shear + self.shear_rate * dt



    def mechanical_relaxation(self,N=1000,cool_factor=0.97,dt=1,theTk=False,StressTensorEvaluator=False):
        """N times repeated `particleShear.CanvasPointsBasicElasticity.mechanical_simulation_step` with graphical
        update

        This method is defined in class `particleShear.CanvasPointsBasicElasticity`"""
        for i in range(N):
            self.mechanical_simulation_step(cool_factor,dt)
            if self.doDrawing:
                theTk.update()
            if(StressTensorEvaluator):
                self.evaluate_stress_tensors(StressTensorEvaluator)

    def set_central_repulsion_coefficient(self,central_repulsion_coefficient=0):
        """Transmit the `particleShear.CanvasPointsBasicElasticity.central_repulsion_coefficient` to the spheres.

        This method is defined in class `particleShear.CanvasPointsBasicElasticity`"""
        self.central_repulsion_coefficient=central_repulsion_coefficient
        for theSphere in self.sphereList:
            theSphere.central_repulsion_coefficient=central_repulsion_coefficient



from particleShearBase import StressTensorEvaluation
from particleShearBase import Force_register
import math

class EvaluationHandler():
    def __init__(self):
        self.strain = []
        self.strain_rate = []
        self.force = []
        self.t = []
        self.shear_stress=[]
        self.Gprime=0
        self.viscosity=0

        self.xg=[0,0] # Center of gravity, in microns
        self.total_mass = 0 # Total mass present, in mg
        self.v_mean=[0,0] # Mean speed, in micrometers/s

        self.angular_momentum_around_origin=0


        self.theEvaluator=False


        # Stress tensors to be stored

        self.stress_tensor_LW = []
        self.stress_tensor_linear_acceleration_otsuki = []
        self.stress_tensor_with_external_forces = []
        self.stress_tensor_linear_acceleration = []
        self.stress_tensor_unbalanced_forces = []
        self.stress_tensor_unbalanced_torque = []
        self.stress_tensor_internal_torque = []
        self.stress_tensor_spin_kinetic_energy = []
        self.overall_stress_tensor = []
        self.overall_stress_tensor2 = []

    def record(self, theEnsemble,theEvaluator=False):
        self.t.append(theEnsemble.t)


        if (not theEvaluator) and (not self.theEvaluator):
            self.theEvaluator=StressTensorEvaluation(theEnsemble.size_x,theEnsemble.size_y)
        if theEvaluator:
            self.theEvaluator=theEvaluator

        # Check whether the paired register is really paired


        sum_x = 0
        sum_y = 0
        total = 0
        for thePair in theEnsemble.force_register.pair_register:
            sum_x = sum_x + thePair[2][0]
            sum_y = sum_y + thePair[2][1]
            total = total + abs(thePair[2][0]+thePair[2][1])

        if total > 0:
            if not (abs(sum_x)/total<1e-14 and abs(sum_y)/total < 1e-14):
                print ("Stress tensor evaluation: Error in force register, paired forces do not match:",sum_x,sum_y)


        # Get the center of gravity:
        xg_x = 0
        xg_y = 0

        vm_x = 0
        vm_y = 0

        self.total_mass=0

        self.angular_momentum_around_origin=0

        for theSphere in theEnsemble.sphereList:
            self.total_mass=self.total_mass+theSphere.m
            xg_x = xg_x + theSphere.x*theSphere.m
            xg_y = xg_y + theSphere.y*theSphere.m

            vm_x = vm_x + theSphere.xspeed*theSphere.m
            vm_y = vm_y + theSphere.yspeed * theSphere.m

            self.angular_momentum_around_origin=self.angular_momentum_around_origin+\
                                                theSphere.m*theSphere.x*theSphere.yspeed-\
                                                theSphere.m*theSphere.y*theSphere.xspeed+\
                                                theSphere.omega*theSphere.inertia

        self.xg=[xg_x/self.total_mass,xg_y/self.total_mass]
        self.v_mean=[vm_x/self.total_mass,vm_y/self.total_mass]





        theEvaluator.evaluate_stress_tensors(theEnsemble.force_register,theEnsemble.movableSphereList(),theEnsemble.shear_rate)

        #print("EvaluationHandler: overall stress tensor ",theEvaluator.overall_stress_tensor)

        self.strain.append(theEnsemble.shear)
        self.strain_rate.append(theEnsemble.shear_rate)


        self.shear_stress.append(theEvaluator.evaluate_quasistatic_shear_stress()) # This is overall shear stress



        self.force.append(theEvaluator.evaluate_externally_applied_shear_stress()) # This is shear stress as theoretically defined by average surface force on
        # arbitrary sections.

        # Stress tensors to be stored
        self.stress_tensor_LW.append(theEvaluator.stress_tensor_LW)
        self.stress_tensor_linear_acceleration_otsuki.append(theEvaluator.stress_tensor_linear_acceleration_otsuki)
        self.stress_tensor_with_external_forces.append(theEvaluator.stress_tensor_with_external_forces)
        self.stress_tensor_linear_acceleration.append(theEvaluator.stress_tensor_linear_acceleration)
        self.stress_tensor_unbalanced_forces.append(theEvaluator.stress_tensor_unbalanced_forces)
        self.stress_tensor_unbalanced_torque.append(theEvaluator.stress_tensor_unbalanced_torque)
        self.stress_tensor_internal_torque.append(theEvaluator.stress_tensor_internal_torque)
        self.stress_tensor_spin_kinetic_energy.append(theEvaluator.stress_tensor_spin_kinetic_energy)
        self.overall_stress_tensor.append(theEvaluator.overall_stress_tensor)








        sum_strain=0
        sum_strain_rate=0

        self.Gprime=0
        self.viscosity=0

        for index in range(len(self.t)):
            sum_strain=sum_strain+self.strain[index]*self.strain[index]
            sum_strain_rate=sum_strain_rate+self.strain_rate[index]*self.strain_rate[index]
            self.Gprime=self.Gprime-self.strain[index]*self.shear_stress[index]
            self.viscosity=self.viscosity-self.strain_rate[index]*self.shear_stress[index]

        if sum_strain > 0:
            self.Gprime=self.Gprime/sum_strain
        if sum_strain_rate > 0:
            self.viscosity=self.viscosity/sum_strain_rate










import unittest
from particleShear import *
from tkinter import *
import math



class TestSingleSphereLinear(unittest.TestCase):

    def setUp(self):

        size_x = 500
        size_y = 500

        theTk = Tk()
        theCanvas = Canvas(theTk, width=size_x, height=size_y)

        theCanvas.pack()

        theEnsemble = EnsembleLinkable(size_x=size_x, size_y=size_y, N=0, packing_fraction=1, theCanvas=theCanvas,
                                       doDrawing=True)

        theEnsemble.sphereList.append(SphereLinkable("GREY", 150, 200, 120, m=1, my_index=1, theCanvas=theCanvas,
                                                     doDrawing=theEnsemble.doDrawing))

        theEnsemble.sphereList[0].xspeed = 0.7
        theEnsemble.sphereList[0].yspeed = 0.3
        # theEnsemble.set_graphical_output_configuration(theEnsemble.graphical_output_configuration)
        theTk.update()

        theEvaluator = StressTensorEvaluation(size_x, size_y)

        self.theEvaluator = theEvaluator
        self.theEnsemble = theEnsemble
        self.theTk=theTk



    def tearDown(self):
        self.theTk.destroy()
        self.theTk = False


    def test_stress_tensor_LW(self):
        for i in range(20):
            self.theEnsemble.mechanical_simulation_step(cool_factor=1, dt=.1)
            self.theTk.update()
            self.theEvaluator.evaluate_stress_tensors(self.theEnsemble.force_register, self.theEnsemble.sphereList, 0)


        LW = self.theEvaluator.stress_tensor_LW
        self.assertTrue(LW[0][0]==0 and LW[0][1]==0 and LW[1][0]==0 and LW[1][1]==0)


    def test_overall_stress_tensor(self):

        for i in range(20):
            self.theEnsemble.mechanical_simulation_step(cool_factor=1, dt=.1)
            self.theTk.update()
            self.theEvaluator.evaluate_stress_tensors(self.theEnsemble.force_register, self.theEnsemble.sphereList, 0)



        overall = self.theEvaluator.overall_stress_tensor
        self.assertTrue(overall[0][0]==0 and overall[0][1]==0 and overall[1][0]==0 and overall[1][1]==0)


    def test_unbalanced_forces_stress_tensor(self):

        for i in range(20):
            self.theEnsemble.mechanical_simulation_step(cool_factor=1, dt=.1)
            self.theTk.update()
            self.theEvaluator.evaluate_stress_tensors(self.theEnsemble.force_register, self.theEnsemble.sphereList, 0)



        unbalanced = self.theEvaluator.stress_tensor_unbalanced_forces
        self.assertTrue(unbalanced[0][0]==0 and unbalanced[0][1]==0 and unbalanced[1][0]==0 and unbalanced[1][1]==0)

    def test_external_stress_tensor(self):

        for i in range(20):
            self.theEnsemble.mechanical_simulation_step(cool_factor=1, dt=.1)
            self.theTk.update()
            self.theEvaluator.evaluate_stress_tensors(self.theEnsemble.force_register, self.theEnsemble.sphereList, 0)

        external = self.theEvaluator.stress_tensor_with_external_forces
        self.assertTrue(
            external[0][0] == 0 and external[0][1] == 0 and external[1][0] == 0 and external[1][1] == 0)


    def test_peculiar_acceleration_stress_tensor(self):

        for i in range(20):
            self.theEnsemble.mechanical_simulation_step(cool_factor=1, dt=.1)
            self.theTk.update()
            self.theEvaluator.evaluate_stress_tensors(self.theEnsemble.force_register, self.theEnsemble.sphereList, 0)

        peculiar = self.theEvaluator.stress_tensor_linear_acceleration_otsuki

        self.assertTrue(
            peculiar[0][0] == 0 and peculiar[0][1] == 0 and peculiar[1][0] == 0 and peculiar[1][1] == 0)

    def test_conservation_angular_momentum(self):

        for i in range(20):
            self.theEnsemble.mechanical_simulation_step(cool_factor=1, dt=.1)
            self.theTk.update()
            self.theEvaluator.evaluate_stress_tensors(self.theEnsemble.force_register, self.theEnsemble.sphereList, 0)

        theHandler = EvaluationHandler()

        theHandler.record(self.theEnsemble, self.theEvaluator)

        old_angular_momentum=theHandler.angular_momentum_around_origin

        for i in range(20):
            self.theEnsemble.mechanical_simulation_step(cool_factor=1, dt=.1)
            self.theTk.update()
            self.theEvaluator.evaluate_stress_tensors(self.theEnsemble.force_register, self.theEnsemble.sphereList, 0)

        theHandler = EvaluationHandler()

        theHandler.record(self.theEnsemble, self.theEvaluator)

        new_angular_momentum=theHandler.angular_momentum_around_origin


        self.assertTrue( abs((old_angular_momentum-new_angular_momentum)/new_angular_momentum)<1e-15)

    def test_conservation_linear_momentum(self):

        for i in range(20):
            self.theEnsemble.mechanical_simulation_step(cool_factor=1, dt=.1)
            self.theTk.update()
            self.theEvaluator.evaluate_stress_tensors(self.theEnsemble.force_register, self.theEnsemble.sphereList, 0)

        theHandler = EvaluationHandler()

        theHandler.record(self.theEnsemble, self.theEvaluator)


        old_linear_momentum=theHandler.total_mass*theHandler.v_mean

        for i in range(20):
            self.theEnsemble.mechanical_simulation_step(cool_factor=1, dt=.1)
            self.theTk.update()
            self.theEvaluator.evaluate_stress_tensors(self.theEnsemble.force_register, self.theEnsemble.sphereList, 0)

        theHandler = EvaluationHandler()

        theHandler.record(self.theEnsemble, self.theEvaluator)

        new_linear_momentum=theHandler.total_mass*theHandler.v_mean


        self.assertTrue( abs((old_linear_momentum[0]-new_linear_momentum[0])/new_linear_momentum[0])<1e-15)

        self.assertTrue(abs((old_linear_momentum[1] - new_linear_momentum[1]) / new_linear_momentum[1]) < 1e-15)


if __name__ == '__main__':
    unittest.main()


from tkinter import *
import math
from .Force_register import Force_register
from .Graphical_output_configuration import Graphical_output_configuration


class CanvasPoints():
    """Canvas for placing objects of type `particleShear.Point` (or subclasses thereof).

    This class maintains a list of objects of type
    `particleShear.Point` or derived and provides basic functions to handle this collection. These objects are generically
    referred to as spheres and are stored in `particleShear.CanvasPoints.sphereList`.\n
    Class defined in subpackage particleShearBase"""
    def __init__(self, size_x, size_y, theCanvas=False, doDrawing=False):
        """Initialize self

        - **parameters**\n
            `size_x` Width of area to be used in pixels = micrometers for the simulation\n
            `size_y` Height of area to be used in pixels = micrometers for the simulation\n
            `theCanvas` possibility to transmit a `tkinter` `Canvas` object for graphical output\n
            `doDrawing` Flag to indicate whether graphical output should be produced or not"""
        self.theCanvas = theCanvas
        """tkinter canvas for drawing the simulation objects"""
        self.size_x = size_x
        """Width of area to be used in pixels = micrometers for the simulation
        
        Generally, but not strictly necessarily, this is also the width of `particleShear.CanvasPoints.theCanvas`"""
        self.size_y = size_y
        """Height of area to be used in pixels = micrometers for the simulation 
        
        Generally, but not strictly necessarily, this is also the height of `particleShear.CanvasPoints.theCanvas`"""
        self.doDrawing = doDrawing
        """Flag to indicate whether graphical output should be produced or not (boolean)"""
        self.sphereList = []
        """list of objects of type `particleShear.Point` or derived. These objects are referred to as spheres"""
        self.graphical_output_configuration=Graphical_output_configuration() # Default graphical output configuration
        """`particleShear.Graphical_output_configuration` to set options how the spheres and their neighboring relations
        are displayed"""
        self.set_graphical_output_configuration(self.graphical_output_configuration) # To propagate it also to the spheres



    def set_graphical_output_configuration(self, graphical_output_configuration):
        """Set new `particleShear.Graphical_output_configuration`

        This method is defined in class `particleShear.CanvasPoints`"""
        self.graphical_output_configuration=graphical_output_configuration
        for theSphere in self.sphereList:
            theSphere.set_graphical_output_configuration(self.graphical_output_configuration)


    def initiateGraphics(self):
        """Initiate drawing of the constituent sphere objects by invokation of
         `particleShear.Circle.initiate_drawing`.

          For this, the objects need to derive from
         `particleShear.Circle` or otherwise provide a initiate_drawing method. \n
          This method is defined in class `particleShear.CanvasPoints`"""
        if(self.doDrawing):
            for theSphere in self.sphereList:
                theSphere.doDrawing=True
                theSphere.initiate_drawing()


    # For recording incoming force reported by the spheres


    def correct_linear_drift(self):
        """Subtract average linear speed from all constituent sphere objects (of type `particleShear.Point`
        or subclasses thereof)

          This method is defined in class `particleShear.CanvasPoints`"""
        l=0
        vx=0
        vy=0
        for theSphere in self.sphereList:
            vx=vx+theSphere.xspeed
            vy=vy+theSphere.yspeed
            l=l+1
        if l>0:
            vx=vx/l
            vy=vy/l
            for theSphere in self.sphereList:
                theSphere.xspeed=theSphere.xspeed-vx
                theSphere.yspeed = theSphere.yspeed - vy


    def particle_info(self):
        """Return information on the type of particle ensemble used. This information
        can for instance be used to properly document output in files"""
        return "CanvasPoints: Geometrical operations only"

    # To be overriden by subclasses with special boundary conditions

    def canMove(self,theSphere):
        """Return boolean to indicate whether a given sphere (object of type `particleShear.Point`
        or subclass thereof) to move on the current canvas\n
        To be overriden by subclasses of `particleShear.CanvasPoints` with special boundary conditions


        This method is defined in class `particleShear.CanvasPoints`"""
        return True



    def do_linear_acceleration(self, dt):
        """Instruct mobile spheres (objects of type `particleShear.Point` or subclasses thereof) to change their speed depending
        on the forces acting on them. The objects need to to derive from `particleShear.CircleMass` or
        otherwise possess a do_linear_acceleration method


        This method is defined in class `particleShear.CanvasPoints`"""
        for theSphere in self.sphereList:
            if self.canMove(theSphere):
                theSphere.do_linear_acceleration(dt)



    def cool(self, f=0.8):
        """Cool the sphere objects (of type `particleShear.Point` or subclass thereof) by a factor of f;
        for f=0 the objects stop, for f=1 this has no
        effect. For this function to work, the objects need to derive from `particleShear.PointLeesEdwards` or
        otherwise present a cool method

        This method is defined in class `particleShear.CanvasPoints`"""
        for theSphere in self.sphereList:
            theSphere.cool(f)

    def move(self, dt=1):
        """Move the spheres (of type `particleShear.Point` or derived)  during a time period dt (in s)

        This method is defined in class `particleShear.CanvasPoints`"""
        for theSphere in self.sphereList:
            theSphere.move(dt)

    def boundary_conditions(self):
        """Invoke `particleShear.Point.boundary_conditions` routine
        on each of the spheres (of type `particleShear.Point` or subclass)

        This method is defined in class `particleShear.CanvasPoints`"""
        for theSphere in self.sphereList:
            theSphere.boundary_conditions(self.size_x, self.size_y)


    def movableSphereList(self):
        """Return the list of mobile spheres (list of objects of type
        `particleShear.Point` or subclasses thereof); for these mobile objects,
        `particleShear.CanvasPoints.canMove` returns True

        This method is defined in class `particleShear.CanvasPoints`"""
        movable_spheres = []
        for theSphere in self.sphereList:
            if (self.canMove(theSphere)):
                movable_spheres.append(theSphere)
        return movable_spheres



    def regular_sphere_distribution(self,rand_fraction=0.5):
        """Distributes the spheres (objects derived from `particleShear.Point` class or subclasses) in a regular
        rectangular grid on the area defined by `particleShear.CanvasPoints.size_x` * `particleShear.CanvasPoints.size_y`

        - **parameters**\n
            `rand_fraction` allows local random fluctuations of the sphere positions around the grid positions\n
            This method is defined in class `particleShear.CanvasPoints`"""
        N_spheres = len(self.sphereList)

        N_rows = int(round(math.sqrt(N_spheres * self.size_y / self.size_x)))


        total_length_to_distribute = N_rows * (self.size_x)

        length_per_sphere = total_length_to_distribute / N_spheres

        y_spacing = self.size_y / (N_rows + 1)

        total_length_to_distribute = total_length_to_distribute + length_per_sphere

        ni = 0

        for theSphere in self.sphereList:
            current_length_pos = length_per_sphere * ni
            row_index = int(math.floor(current_length_pos / total_length_to_distribute * N_rows))
            col_index = \
                (current_length_pos - row_index * total_length_to_distribute / N_rows) / length_per_sphere
            theSphere.y = row_index * length_per_sphere+(random.random()-0.5)*rand_fraction*length_per_sphere
            theSphere.x = col_index * y_spacing+(random.random()-0.5)*rand_fraction*y_spacing

            ni = ni + 1







from particleShearLinkableObjects import EnsembleCompactParticlesFromModelParameters
from particleShearObjects import particle_shear_model_parameters
from .OscillatoryShearExperiment import OscillatoryShearExperiment
import math
from tkinter import *

from .OscillatorySimulationFragmentation import OscillatorySimulationFragmentation



class Simulation_dermal_filler_rheology(OscillatorySimulationFragmentation):

    def __init__(self,root_folder,do_permanent_links=True,cut_lines=5,N=150,packing_fraction=2,
                 bimodal_factor=1.4,
                 amplitude=0.5,
                 Young_modulus_spheres=1000,
                 density=50,
                 mu=0.1,relative_viscosity=0.01,
                 relative_y_scale_force=5e-4,relative_frequency=0.05, relative_transversal_link_strength=1,
                 central_repulsion_coefficient=1,
                 avoid_horizontal_angle_degree=0,avoid_height_spanning_particles=False,
                 doCutByTriangulation=True,doDrawing=True,saveData=False):

        self.avoid_horizontal_angle_degree=avoid_horizontal_angle_degree
        self.avoid_height_spanning_particles=avoid_height_spanning_particles
        self.doCutByTriangulation=doCutByTriangulation
        self.bimodal_factor=bimodal_factor
        self.central_repulsion_coefficient=central_repulsion_coefficient
        self.density=density

        super(Simulation_dermal_filler_rheology,self).__init__(
            root_folder,do_permanent_links=do_permanent_links,cut_lines=cut_lines,
            N=N,packing_fraction=packing_fraction,amplitude=amplitude,
                 Young_modulus_spheres=Young_modulus_spheres,mu=mu,
                 relative_y_scale_force=relative_y_scale_force,relative_frequency=relative_frequency,
                 doDrawing=doDrawing,saveData=saveData)

        self.relative_viscosity=relative_viscosity
        self.relative_transversal_link_strength=relative_transversal_link_strength

        self.function_call = "Simulation_dermal_filler_rheology(root_folder=" + str(root_folder) + ",do_permanent_links=" + \
                             str(do_permanent_links) + ",cut_lines=" + str(cut_lines) + ",N=" + str(N) + \
                             ",packing_fraction=" + str(packing_fraction) + \
                             ",bimodal_factor="+str(bimodal_factor)+\
                             ",amplitude=" + str(amplitude) + \
                             ",Young_modulus_spheres=" + str(Young_modulus_spheres) + \
                             ",density=" + str(density) + ",mu=" + str(mu) +  \
                             ",relative_viscosity="+str(relative_viscosity)+\
                             ",relative_y_scale_force=" + str(relative_y_scale_force) + ",relative_frequency=" + \
                             str(relative_frequency) +\
                             ",relative_transversal_link_strength="+str(relative_transversal_link_strength)+ \
                             ",central_repulsion_coefficient=" + str(central_repulsion_coefficient) + \
                             ",avoid_horizontal_angle_degree="+str(avoid_horizontal_angle_degree)+\
                             ",avoid_height_spanning_particles=" + str(self.avoid_height_spanning_particles) + \
                             ",doCutByTriangulation="+str(self.doCutByTriangulation) + \
                             ",doDrawing=" + str(doDrawing) + ")"




    def adjustment_factor_for_bounded_preequilibration_steps(self):
        f=1
        if self.mu<=0.1:
            f=f*1.5
        if self.mu<=0.01:
            f=f*2
        if self.mu<=0.001:
            f=f*2
        if self.amplitude<=0.01:
            f=f*(1-(math.log10(self.amplitude)+2))
        if self.amplitude<=0.005:
            f=f*2
        if self.relative_viscosity<=0.1:
            f=f*2
        print("Adjustment factor of pre-equilibration time for low amplitude and friction:",f)
        return int(f)


    def initEnsemble(self,do_debug=False):
        super(Simulation_dermal_filler_rheology,self).initEnsemble()

        self.theEnsemble = EnsembleCompactParticlesFromModelParameters(
            cut_lines=self.cut_lines,N=self.N,packing_fraction=self.packing_fraction,
            Young_modulus_spheres=self.Young_modulus_spheres,
            density=self.density,bimodal_factor=self.bimodal_factor,
            do_permanent_links=self.do_permanent_links,mu=self.mu,theTk=self.theTkSimulation,do_pre_equilibration=True,
            relative_viscosity=self.relative_viscosity,central_repulsion_coefficient=self.central_repulsion_coefficient,
            anticipated_amplitude=self.amplitude,relative_transversal_link_strength=self.relative_transversal_link_strength,
            avoid_horizontal_angle_degree=self.avoid_horizontal_angle_degree,
            avoid_height_spanning_particles=self.avoid_height_spanning_particles,
            doCutByTriangulation=self.doCutByTriangulation,doDrawing=self.doDrawing,
            do_debug=do_debug)

        self.dt=self.theEnsemble.dt
        self.dt_max = self.theEnsemble.dt_max













from tkinter import *
import random
import math
from .Sphere import Sphere
from .SphereFriction import SphereFriction
from particleShearBase import Force_register
from particleShearBase import Graphical_output_configuration
from particleShearBase import CanvasPointsBasicElasticity


def r_estimate(size_x,size_y,packing_fraction,N,bimodal_upper=1.2,bimodal_lower=0.8):
    """Average radius

    TWe find the average radius from the `packing_fraction`, the available area
    `size_x` times `size_y` and the desired number of spheres. Indeed, collectively the `N` spheres should
    occupy a theoretical area of `size_x` times `size_y` divided by `N`. In a real situation, they will of
     course occupy
    a smaller area due to compression. However, as the `packing_fraction`, also referred to as notional phase volume
     (Evans, I.D. and A. Lips, Concentration-Dependence of the Linear Elastic Behavior of Model Microgel Dispersions.
     Journal of the Chemical Society-Faraday Transactions, 1990. 86(20): p. 3413-3417.) is defined relative to the fully
     expanded state, this does not matter. If the spheres had a homogeneous size, the relevant radius would be \n
     sqrt(size_x * size_y * packing_fraction / N / math.pi) \n

    Now, to avoid crystallization, a bimodal distribution of sphere radii needs to be used
    Otsuki, M. and H. Hayakawa, Discontinuous
    change of shear modulus for frictional
    jammed granular materials. Phys Rev E, 2017. 95(6-1): p. 062902. Still, the area occupied by the spheres should
    match size_x * size_y * packing_fraction. In analogy to Otsuki et al.,
    We chose our bimodal distribution at radii +/- 20 percent of the nominal radius, as the sum of the areas of a
    circle of 1.2 times r + the area of a circle of 0.8 times r is slightly bigger than twice the area of a circle of
    radius r, we correct for this. The correction is however minor, its just 4 percent.
    """
    return math.sqrt(size_x * size_y * packing_fraction / N / math.pi / (bimodal_upper * bimodal_upper + \
                                                                         bimodal_lower * bimodal_lower) * 2)


def adjust_m(m,r_actual,r_mean):
    """Adjust the mass for the actual radius as compared to the average theoretical radius"""
    return m*r_actual*r_actual/r_mean/r_mean

class Ensemble(CanvasPointsBasicElasticity):
    """Ensemble of particles of type  `particleShear.SphereFriction`, with regular boundary conditions

        This class creates and maintains a list of objects of type
        `particleShear.SphereFriction`. These objects are generically
        referred to as spheres and are stored in `particleShear.CanvasPoints.sphereList`.\n
        The ensemble class does not implement proper Lees-Edwards boundaries, but emulates neighboring spheres by
        creating replicates of the spheres near the boundares displaced by one positive or negative multiple of the
        xy dimensions of the simulation area.\n
        Class defined in subpackage particleShearObjects"""

    def __init__(self, size_x, size_y, N, packing_fraction=0.8, theCanvas=False, doDrawing=False, k=1, nu=0.01,m=1,
                 bimodal_factor=1.4):
        """Initialize self

        - **parameters**\n
        `size_x` Width of area to be used in pixels = micrometers for the simulation\n
        `size_y` Height of area to be used in pixels = micrometers for the simulation\n
        `N` The number of spheres to place
        `packing_fraction` The packing fraction indicating the density; this is defined as the area occupied by the
        non-compressed spheres as compared to the actual available area (size_x times size_y)
        `theCanvas` possibility to transmit a `tkinter` `Canvas` object for graphical output\n
        `doDrawing` Flag to indicate whether graphical output should be produced or not
        `k` Central spring constant in (mg/s^2)/m of depth;
        so F=-k*x*L with L the depth of the stack (the simulation is 2D) and x the compression\n
        `nu` Central viscosity force constant in (mg/s)/m of depth. This is to have F=nu*v*L, L again the depth\n
        `m` is the mass per unit of depth in mg/m of depth\n
        `bimodal_factor` is ratio of radii of the smaller and larger spheres in the bimodal distribution\n
        """

        super(Ensemble,self).__init__(size_x=size_x,size_y=size_y,theCanvas=theCanvas,doDrawing=doDrawing,k=k, nu=nu,m=m)

        self.bimodal_upper = math.sqrt(bimodal_factor)
        """Upper multiplication factor for the bimodal sphere size distribution"""
        self.bimodal_lower = math.sqrt(1/bimodal_factor)
        """Lower multiplication factor for the bimodal sphere size distribution"""

        self.N = N
        """Number of spheres"""

        self.packing_fraction = packing_fraction
        """Packing fraction: relative area occupied by the uncompressed spheres as compared to the available area"""

        self.k = k
        """Central spring constant in (mg/s^2)/m of depth"""
        self.nu = nu
        """Central viscosity constant in (mg/s)/m of depth"""


        self.record_type = "internal"
        """Helper variable to distinguish between internal spheres and boundary replicates for the accounting of internal
        and external forces"""


        r=self.size_y/2 # If there are no spheres specified, keep some default value

        if N>0:
            r = r_estimate(self.size_x , self.size_y ,packing_fraction , N,
                           bimodal_upper=self.bimodal_upper,
                           bimodal_lower=self.bimodal_lower )

        self.r_max = 0
        """Largest sphere radius"""
        for i in range(N):
            r_sphere = self.random_radius_bimodal(r)
            if r_sphere > self.r_max:
                self.r_max = r_sphere
            self.sphereList.append(SphereFriction("grey",
                                                     random.randrange(0, size_x),
                                                     random.randrange(0, size_y), r_sphere * 2,
                                                     adjust_m(m, r_sphere, r),
                                                     i,theCanvas, False,self))
        self.spheres_boundary = []
        """List of the spheres on the boundary"""

        self.set_graphical_output_configuration(self.graphical_output_configuration) # To propagate it also to the spheres

        self.initiateGraphics()





    def random_radius_bimodal(self,r):
        """Choose randomly one of the two bimodal radius values.

        Method defined in `particleShear.Ensemble`"""
        return r*(self.bimodal_lower+(self.bimodal_upper-self.bimodal_lower)*random.randint(0, 1))


    # For recording incoming force reported by the spheres

    def record_individual_internal_force(self,target,source,force_vector):
        """Record internal force

        Internal forces are the forces that act between spheres within the interior of the simulation. That is,
        they are able to move, and are not replicates to emulate periodic or other boundaries. In this class, the
        distinction between internal and external forces is made by using the instance variable `particleShear.Ensemble.record_type`
        which switches between internal and external depending on whether forces originating from internal spheres
        or boundary replicates are considered.\n\n
        Method defined in `particleShear.Ensemble`"""

        if not self.canMove(target):
            return
        if self.record_type=="internal":
            super(Ensemble,self).record_individual_internal_force(target,source,force_vector)
        else: # running virtual boundary replicates, this is also external here
            self.record_external_force(target, force_vector)




    def particle_info(self):
        """Provide short description of particle type of inclusion into output file

        Method defined in `particleShear.Ensemble` """
        return "Non-frictional spheres, \n\tbimodal distribution, equal probability, \n\tr1="+\
               str(self.r_max*0.8/1.2)+"r2="+str(self.r_max)+"\n\tidentical mass="+str(self.sphereList[0].m)+"mg"

    def test_neighbor_relation(self):
        """Instruct every sphere to test establish its geometric (and possibly permanent) neighbors.

        This function calls `particleShear.CircleMassNeighbors.test_neighbor_relation` on each member of the
        `particleShear.CanvasPoints.sphereList`. It does so not only on the actuale sphere list, but also on the
        replicates used to emulate the boundary environment in this class.\n
        This method is defined in class `particleShear.Ensemble`"""
        super(Ensemble,self).test_neighbor_relation()
        for sphereIndex in range(len(self.sphereList)):
             for sphereIndexBoundary in range(len(self.spheres_boundary)):
                self.sphereList[sphereIndex].test_neighbor_relation(self.spheres_boundary[sphereIndexBoundary])


    def elastic_force(self):
        """Let all spheres calculate the central elastic force acting on them

        This includes forces acting from the replicated boundary spheres.

        This method is defined in class `particleShear.Ensemble`"""
        self.record_type = "interal"
        super(Ensemble,self).elastic_force()
        self.record_type = "external"
        for sphereIndex in range(len(self.sphereList)):
             for sphereIndexBoundary in range(len(self.spheres_boundary)):
                self.sphereList[sphereIndex].elastic_force(self.spheres_boundary[sphereIndexBoundary], self.k)


    def central_viscous_force(self):
        """Let all spheres calculate the central elastic force acting on from their known neighbors.

         This includes forces acting from the replicated boundary spheres. \n
         This method is defined in class `particleShear.CanvasPointsBasicElasticity`"""

        self.record_type = "interal"
        super(Ensemble, self).central_viscous_force()
        self.record_type = "external"
        for sphereIndex in range(len(self.sphereList)):
              for sphereIndexBoundary in range(len(self.spheres_boundary)):
                self.sphereList[sphereIndex].central_viscous_force(self.spheres_boundary[sphereIndexBoundary],
                                                                   self.nu)



    def periodic_boundary_extension(self):
        """Create periodic boundary extension.

         The function replicates spheres near the boundary to obtain virtual copies
         displaced by one positive or negative of the simulation area, in both x and y direction and also
         in corners by displacement in both x and y direction.


         This method is defined in class `particleShear.Ensemble`"""

        if self.doDrawing:

            for theSphere in self.spheres_boundary:
                theSphere.deleteDrawing()

        max_index=0
        for theSphere in self.sphereList:
            if max_index<theSphere.myindex:
                max_index=theSphere.myindex

        self.spheres_boundary = []
        index=max_index+1
        for theSphere in self.sphereList:
            pos = theSphere.coordinates()

            left_edge = pos[0] - pos[2] <= self.r_max
            right_edge = pos[0] + pos[2] >= self.size_x - self.r_max
            upper_edge = pos[1] - pos[2] <= self.r_max
            lower_edge = pos[1] + pos[2] >= self.size_y - self.r_max

            if left_edge:
                self.spheres_boundary.append(SphereFriction(
                    "green", pos[0] + self.size_x, pos[1], 2 * pos[2],
                    theSphere.m,index,self.theCanvas, self.doDrawing))
                index=index+1
                if upper_edge:
                    self.spheres_boundary.append(SphereFriction(
                        "green", pos[0] + self.size_x, pos[1] + self.size_y, 2 * pos[2],
                        theSphere.m,index, self.theCanvas, self.doDrawing))
                    index=index+1
                if lower_edge:
                    self.spheres_boundary.append(SphereFriction(
                        "green", pos[0] + self.size_x, pos[1] - self.size_y, 2 * pos[2],
                        theSphere.m,index, self.theCanvas, self.doDrawing))
                    index=index+1
            if right_edge:
                self.spheres_boundary.append(SphereFriction(
                    "green", pos[0] - self.size_x, pos[1], 2 * pos[2],
                    theSphere.m,index, self.theCanvas, self.doDrawing))
                index=index+1
                if upper_edge:
                    self.spheres_boundary.append(SphereFriction(
                        "green", pos[0] - self.size_x, pos[1] + self.size_y, 2 * pos[2],
                        theSphere.m,index,self.theCanvas, self.doDrawing))
                    index=index+1
                if lower_edge:
                    self.spheres_boundary.append(SphereFriction(
                        "green", pos[0] - self.size_x, pos[1] - self.size_y, 2 * pos[2],
                        theSphere.m, index,self.theCanvas, self.doDrawing))
                    index=index+1
            if upper_edge:
                self.spheres_boundary.append(SphereFriction(
                    "green", pos[0], pos[1] + self.size_y, 2 * pos[2],
                    theSphere.m,index, self.theCanvas, self.doDrawing))
                index=index+1
            if lower_edge:
                self.spheres_boundary.append(SphereFriction(
                    "green", pos[0], pos[1] - self.size_y, 2 * pos[2],
                    theSphere.m,index, self.theCanvas, self.doDrawing))
                index=index+1


    def mechanical_simulation_step_calculate_forces(self):
        """ Calculate the forces acting on the spheres

        This method is defined in class `particleShear.Ensemble`"""

        self.reset_force_register()
        self.periodic_boundary_extension()
        self.test_neighbor_relation()
        self.elastic_force_from_neighbors()
        self.central_viscous_force_from_neighbors()







# High level functions to run simulations in a standardized way


__all__ = [
            "doParticleShearSimulation",
"doParticleShearSimulationSeries"]








# Simulation stuff
from .doParticleShearSimulation import doParticleShearSimulation
from .doParticleShearSimulationSeries import doParticleShearSimulationSeries


from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(name='particleShear',
      version='1.0.1',
      description='Simulation of frictional and possibly crosslinked particles in shear',
      url='https://github.com/tbgitoo/particleShear',
      author='Thomas Braschler',
      author_email='thomas.braschler@unige.ch',
      long_description=long_description,
      long_description_content_type="text/markdown",
      license='MIT',
      packages=['particleShear','particleShearBase','particleShearObjects',
                'particleShearLinkableObjects',
                'particleShearSimulation',
                'particleShearRunSimulation',
                'particleShearTest'],
      install_requires=['Pillow'],
      zip_safe=False,
      test_suite='particleShearTest',
      project_urls={
          "Bug Tracker": "https://github.com/tbgitoo/particleShear/issues"
      },
      classifiers=[
          "Development Status :: 4 - Beta",
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
          "Operating System :: OS Independent",
      ],
    python_requires='>=3.5'
      )